/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * Event Loop HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : evt-loop.h
 * CREATION DATE : Mar 30, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/
#ifndef _EVENT_EVT_LOOP_H_
#define _EVENT_EVT_LOOP_H_

#include <nmapp/types/nm-types.h>
#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include <nmapp/logging/ILog.h>
#include <nmapp/container/uthash.h>
#include <nmapp/sys/IUnixInitSys.h>



/** @ingroup libnmapp
 * @defgroup evt-loop Event Loop
 * Provides a C API using libevent for writing simple event based applications. The libevent
 * bases on select (see man select).
 */
#ifdef __cplusplus
extern "C" {
#endif


/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/** Forward declaration of the event loop object reference */
typedef struct __EvtLoopPrivate      * EvtLoopRef;

/** Forward declaration of the event loop event object reference */
typedef struct __EvtLoopEventPrivate * EvtLoopEventRef;

/** Forward declaration of the event loop timer object reference */
typedef struct __EvtLoopTimerPrivate * EvtLoopTimerRef;

/** Event loop callback function pointer prototype */
typedef void (*EvtLoopEventCallback)(void *pCtx);          //!< Event loop event callback function pointer


/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Create Event-Loop Instance
 * Creates an event loop instance where also basic exit signals added
 *
 * @param logRef          A reference to the logger interface to use
 * @param useExitSignals  true = exit signals will be created and added
 * @return A reference to the created instance
 */
EvtLoopRef EvtLoop_create(ILogRef logRef, bool useExitSignals);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Destroy Event-Loop Instance
 * Destroys an event loop instance
 *
 * @param me    A reference to an event loop instance
 */
void EvtLoop_destroy(EvtLoopRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Re-Iniitalize an Event-Loop instance
 * Reinitializes the event loop instance
 *
 * @param me    A reference to the event loop
 */
void EvtLoop_reinit(EvtLoopRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Run an Event-Loop instance
 * Runs an event loop instance
 *
 * @param me    A reference to the event loop
 * @return The return value when the event loop exits
 */
int EvtLoop_run(EvtLoopRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Quit an Event-Loop instance
 * Quits a running event loop instance
 *
 * @param me    A reference to the event loop
 */
void EvtLoop_quit(EvtLoopRef me);


//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Register a Signal Event to the Event-Loop
 * Registers a signal to the event loop
 *
 * @param evtLoopRef    A reference to the event loop
 * @param signalNbr     The number of the signal
 * @param cbFn          A pointer to the signal callback function
 * @param pCtx          A pointer to the user context
 */
void EvtLoop_registerSignalEvent(EvtLoopRef evtLoopRef, int signalNbr, EvtLoopEventCallback cbFn, void *pCtx);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Unregister a Signal Event from the Event-Loop
 * Unregisters a signal from the event loop
 *
 * @param evtLoopRef    A reference to the event loop
 * @param signalNbr     The number of the signal
 */
void EvtLoop_unregisterSignalEvent(EvtLoopRef evtLoopRef, int signalNbr);


//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Register a EPoll Event to the Event-Loop
 * Registers an epoll event to the event loop
 *
 * @param evtLoopRef      A reference to the event loop
 * @param fd              The file descriptor of the file
 * @param cbFn            A pointer to the file changed callback function
 * @param isRegularFile   True = regular file shall be epolled, False = linux exception event
 * @param pCtx            A pointer to the user context
 */
void EvtLoop_registerEPollEvent(EvtLoopRef evtLoopRef, int fd, EvtLoopEventCallback cbFn, bool isRegularFile, void *pCtx);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Unregister a EPoll Event from the Event-Loop
 * Unregisters and removes an epoll event from the event loop
 *
 * @param evtLoopRef      A reference to the event loop
 * @param fd              The file descriptor of the file
 */
void EvtLoop_unregisterEPollEvent(EvtLoopRef evtLoopRef, int fd);


//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Register a Filesystem Change Event to the Event-Loop
 * Registers an inotify file descriptor to the event loop for monitoring file system changes.
 *
 * @param evtLoopRef      A reference to the event loop
 * @param filePath        The pathname of the file to register
 * @param cbFn            A pointer to the file changed callback function
 * @param pCtx            A pointer to the user context
 */
void EvtLoop_registerFilesystemChangeEvent(EvtLoopRef evtLoopRef, const char* filePath, EvtLoopEventCallback cbFn, void *pCtx);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Unregister a Filesystem Change Event from the Event-Loop
 * Unregisters and removes an inotify file descriptor from the event loop
 *
 * @param evtLoopRef      A reference to the event loop
 * @param filePath        The pathname of the file to register
 */
void EvtLoop_unregisterFileSystemChangeEvent(EvtLoopRef evtLoopRef, const char* filePath);


//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Create and Register an Event-Loop timer
 * Creates and register a timer event to the event loop
 *
 * @param evtLoopRef    A reference to the event loop
 * @param timeoutCbFn   A pointer to the timeout callback function
 * @param timeoutMS     The timeout [ms]
 * @param isPeriodic    true = periodic timer, false = one shot timer
 * @param pCtx          A pointer to the user context
 * @return The reference to the created timer
 */
EvtLoopTimerRef EvtLoop_createAndRegisterTimer(EvtLoopRef evtLoopRef, EvtLoopEventCallback timeoutCbFn,
                                               uint32_t timeoutMS, bool isPeriodic, void *pCtx);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Destroy an Event-Loop Timer instance
 * Destroys an timer instance
 *
 * @param me    A reference to the time event loop
 */
void EvtLoop_destroyTimer(EvtLoopTimerRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Start the Event-Loop Timer
 * Starts the timer of an event loop timer instance
 *
 * @param me    A reference to the event loop timer
 */
void EvtLoop_startTimer(EvtLoopTimerRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Stop the Event-Loop Timer
 * Stops the timer of an event loop timer instance
 *
 * @param me    A reference to the event loop timer
 */
void EvtLoop_stopTimer(EvtLoopTimerRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Is Timer Running
 * Returns whether timer is running or not
 *
 * @param me    A reference to the event loop timer
 * @return true = timer is active and running, false otherwise
 */
bool EvtLoop_isTimerRunning(EvtLoopTimerRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Set Timeout of an Event-Loop Timer
 * Sets the timeout of an event loop timer instance
 *
 * @param me         A reference to the event loop timer
 * @param timeoutMS  The timeout in ms
 */
void EvtLoop_setTimerTimeout(EvtLoopTimerRef me, uint32_t timeoutMS);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Get Timeout of an Event-Loop Timer
 * Gets the timeout of an event loop timer instance
 *
 * @param me         A reference to the event loop timer
 * @return The timeout in ms
 */
uint32_t EvtLoop_getTimerTimeout(EvtLoopTimerRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Set Periodicity of an Event-Loop Timer
 * Sets the periodicity of an event loop timer instance
 *
 * @param me         A reference to the event loop timer
 * @param isPeriodic true = periodic, false = one-shot event
 */
void EvtLoop_setTimerPeriodicity(EvtLoopTimerRef me, bool isPeriodic);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Is an Event-Loop Timer Periodic
 * Returns whether an event loop timer instance is periodic
 *
 * @param me         A reference to the event loop timer
 * @return true = is periodic, false = one-shot timer
 */
bool EvtLoop_isTimerPeriodic(EvtLoopTimerRef me);


//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Register a DBus Signal to the Event-Loop
 * Registers a dbus signal to the event loop
 *
 * @param evtLoopRef     A reference to the event loop
 * @param initSysItfRef  A reference to the init system interface
 * @param dbusService    The dbus service (string), e.g. "org.freedesktop.systemd1"
 * @param dbusObjPath    The dbus object path, e.g. "/org/freedesktop/systemd1"
 * @param dbusItfName    The dbus interface name, e.g. "org.freedesktop.systemd1.Manager"
 * @param dbusSignalName The sbus signal name, e.g. "UnitNew"
 * @param dbusSignalItem The sbus signal item, e.g. "poweroff.target"
 * @param cbFn           A pointer to the signal item callback function
 * @param pCtx           A pointer to the user context
 */
void EvtLoop_registerDBusSignal(EvtLoopRef evtLoopRef, IUnixInitSysRef initSysItfRef, const char* dbusService,
                                const char* dbusObjPath, const char* dbusItfName, const char* dbusSignalName,
                                const char* dbusSignalItem, EvtLoopEventCallback cbFn, void *pCtx);
//--------------------------------------------------------------------------------------------
/** @ingroup evt-loop
 * @brief Unregister a DBus Signal from the Event-Loop
 * Unregisters a dbus signal from the event loop
 *
 * @param evtLoopRef     A reference to the event loop
 * @param initSysItfRef  A reference to the init system interface
 * @param dbusService    The dbus service (string), e.g. "org.freedesktop.systemd1"
 * @param dbusObjPath    The dbus object path, e.g. "/org/freedesktop/systemd1"
 * @param dbusItfName    The dbus interface name, e.g. "org.freedesktop.systemd1.Manager"
 * @param dbusSignalName The dbus signal name, e.g. "UnitNew"
 * @param dbusSignalItem The sbus signal item, e.g. "poweroff.target"
 */
void EvtLoop_unregisterDBusSignal(EvtLoopRef evtLoopRef, IUnixInitSysRef initSysItfRef, const char* dbusService,
                                  const char* dbusObjPath, const char* dbusItfName, const char* dbusSignalName,
                                  const char* dbusSignalItem);




#ifdef __cplusplus
}
#endif

#endif /* INCLUDE_NMAPP_EVENT_EVT_LOOP_H_ */
/************************************END OF FILE*********************************************/
