/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * Exception Handler HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : exception.h
 * CREATION DATE : Mar 10, 2020
 * TARGET        : generic
 *
 *********************************************************************************************/
#ifndef _EXCEPTION_H_
#define _EXCEPTION_H_

#ifdef __cplusplus
extern "C"
{
#endif

/**
* \ingroup libnm
* \defgroup exceptionHdl Exception Handler
* The exception handler is implemented in one c module:
* - exception.h
*   - declaration of the exception type \ref Exception_t
*   - declaration of all known exceptions \ref ExcpetionId_t
*   - provides the \b Try-Throw-Catch-Finally functionality using macros. \e
*   .
* - exception.c
*   - static array of exception instances (min. one per thread using the exception handler)
*   - defines the platform specific getExceptionInstance functions (determining the instance for a calling thread)
* .
* The implementation files (c-files) must only include 'exception.h'.\n
*
* \b \e Rethrowing/Not-Rethrowing: In case of nested Try,Throw,Catch statements the innermost Catch-statement handles the
*                                  thrown exception as first. You can also rethrow the exception if needed.\n\n
*
* \section references References
* Googling something like 'exception handler in ANSI C' resulted in URLs like:
* - https://code.google.com/p/exceptions4c/
* - http://code.google.com/p/libex/
* - http://www.di.unipi.it/~nids/docs/longjump_try_trow_catch.html
* - http://www.nicemice.net/cexcept/
* .
* Most of these references share a common mechanism: setjmp.h (setjmp and longjmp), see \ref setjmpFunction for details.\n
* To start with a simple exception handler, latter two URLs were taken as base.\n\n
*
* \section exceptionHdlUsage Exception Handler Usage
* The following subsections show how to use the exception handler.\n
*
* \subsection exceptionHdlGeneralUsage General Usage
* Generally the machanism follows the \e try-throw-catch mechanism similar to object-oriented languages.\n
* The following example shows the \b Try-Catch construct. Each statement within \b Try is able to \e throw an exception.\n
\verbatim
 Exception_t e;

 Try
 {
     // statement 1
     // statement 2
     // ...
 }
 Catch(e)
 {
     switch (e.id)
     {
         case EX_OVERFLOW:                     // see enum ExcpetionId_t
             // handle overflow denied exception
             break;
         // ...
         default:
             // unknown exception
             break;
     }
 }
\endverbatim
* \n
*
*
* \subsection exceptionHdlSetup How to Setup the Exception Handler
* There is no need for a specific setup anymore. Each "Try" macro will determine it's own exception instance and
* initialize it correctly.
* \n
*
* \subsection throwUsage Throw Usage
* \ref exceptionHdlGeneralUsage describes how to use the \b Try-Catch construct, where each statement under \b Try can \b Throw an
* exception. Throwing an exception is illustrated in the following code snippet.\n
\verbatim
 void statementFunction(int a)
 {
     if (a > 31)
     {
         Throw(EX_LIMIT_CHECK, "Exceeding limit by a (%d)", a);
     }
}
\endverbatim
* \n
*
* \section multiThread Multi-Threading
* If an application using \e exception.h creates a thread, each thread needs it's own exception instance. Therefore
* the Try, Catch and Throw macros are using a platform specific mechanism to determine the id of the calling thread
* and will use it to identify the correct instance.
*
* \n
*
* \section setjmpFunction How does setjmp work
* \b setjmp stores the stack content/environment in a buffer called jmp_buf. When calling \b setjmp \e directly it returns 0.
* \b longjmp restores the stack content/environment stored at \b setjmp and holds a return value by its second argument.\n
* After \b longjmp is completed, program execution continuous as if the corresponding call of \b setjmp (given by the jmp_buf
* variable) had just returned the value that \b longjmp provides with its second argument.\n
* \n
*
*/



/*
**********************************************************************************************
*                   INCLUDES
**********************************************************************************************
*/
#include <setjmp.h>
#include <stdio.h>

#include <nmapp/types/nm-types.h>

/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/**
 * Footprint definitions
 */
#define EXCEPTION_MAX_INSTANCES         16      // Maximum number of independent exception instances
#define EXCEPTION_MSG_SIZE_MAX         256      // Max exception message size, see field msg of \ref Exception_t


/**
* Enumerated declaration of all exceptions
* The following exception id's are generic and are unique identifiers used by all applications.
* ATTENTION: don't change the numbers.
*/
typedef enum __exceptionId
{
    EX_INVALID_ARG          =    1,     //!< Invalid argument
    EX_INIT                 =    2,     //!< Initialization e.g. a mutex
    EX_CREATE               =    3,     //!< Creating e.g. a file
    EX_DESTROY              =    4,     //!< Destroyin e.g. a file
    EX_OPEN                 =    5,     //!< Opening e.g. a file
    EX_CLOSE                =    6,     //!< Closing e.g. a file
    EX_ADD                  =    7,     //!< Adding e.g. an item
    EX_REMOVE               =    8,     //!< Removing e.g. an item
    EX_WRITE                =    9,     //!< Writing e.g. an item
    EX_READ                 =   10,     //!< Reading e.g. an item
    EX_LOCK                 =   11,     //!< Locking e.g. an item
    EX_UNLOCK               =   12,     //!< Unlocking e.g. an item
    EX_SUBSCRIBE            =   13,     //!< Subscribing, e.g. at an observer
    EX_UNSUBSCRIBE          =   14,     //!< Unsubscribing, e.g. from an observer
    EX_MISSING              =   15,     //!< Missing/Not existing file, directory, etc.
    EX_MISMATCH             =   16,     //!< Mismatching, e.g. type mismatch
    EX_NOT_FOUND            =   17,     //!< the search element was not found
    EX_UNDEFINED            =   18,     //!< Undefined, e.g. undefined state
    EX_UNSUPPORTED          =   19,     //!< Unsupported, e.g. exit application from io-svc
    EX_TRANSMIT             =   20,     //!< Transmitting
    EX_RECEIVE              =   21,     //!< Receiving
    EX_OUT_OF_RANGE         =   22,     //!< Parameter / value is out of range
    EX_OUT_OF_MEMORY        =   23,     //!< out of memory exception (e.g. malloc failed)
    EX_OVERFLOW             =   24,     //!< overflow (e.g. on memory buffer write access)
    EX_LIMIT_CHECK          =   25,     //!< Limit check, e.g. check if message queue is full
    EX_CANCEL               =   26,     //!< Cancelling, e.g. a thread
    EX_JOIN                 =   27,     //!< Joining, e.g. a thread
    EX_START                =   28,     //!< Starting, e.g. a timer
    EX_STOP                 =   29,     //!< Stopping, e.g. a timer
    EX_INACTIVE             =   30,     //!< Inactive, e.g. at parsing for an active event
    EX_CONVERSION           =   31,     //!< Conversion failure (e.g. string to float)
    EX_TIMEOUT              =   32,     //!< Timeout
    EX_MOUNT                =   33,     //!< Mounting e.g. a device
    EX_UNMOUNT              =   34,     //!< Unmounting e.g. a device
    EX_SYNC                 =   35,     //!< Synchronization, e.g. thread sync
    EX_MALFORMED            =   36,     //!< Malformed string, file, configuration
    EX_FORK                 =   37,     //!< Forking error
    EX_VALIDATION           =   38,     //!< Validation error
    EX_UPDATE_FAILURE       =   39,     //!< Update failed
    EX_NACK                 =   40,     //!< Not acknowledged, e.g. message command
    EX_CHECKSUM             =   41,     //!< Checksum error
    EX_NOT_CONNECTED        =   42,     //!< connection missing, e.g. tcp-client socket (send or receive failed)

    /** End of exception declaration (do not add any exception after this one!) */
    EX_MAX_EXCEPTIONS       = 0xFFFF    //!< Max number of exceptions
} ExcpetionId_t;


/** Content declaration of an exception type */
typedef struct __Exception
{
    ExcpetionId_t        id;                               //!< Identification of the exception, i.e. one selection out of \ref ExcpetionId_t
    volatile const char *file;                             //!< File in which the exception is thrown
    volatile const char *function;                         //!< Function in which the exception is thrown
    unsigned int         line;                             //!< Line of the exception source
    char                 msg[EXCEPTION_MSG_SIZE_MAX];      //!< Message belonging to the exception
} Exception_t;


/** Thread reference definition */
typedef struct __ThreadStruct * threadRef;


/** Structure definition of an exception instance */
typedef struct __ExceptionInstance
{
    jmp_buf *                 pStack;
    volatile char             caught;
    volatile struct           { Exception_t current; } exception;
    bool                      isUsed;
    threadRef                 thread;
    int32_t                   numbOfNestedTrys;
} ExceptionInstance_t;



/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/** @ingroup exceptionHdl
 * @brief Get Exception instance
 * Returns an newly created or existing exception instance depending on the thread ID.
 * \e NOTE: Can end in an assertion if all exception instances are used
 *
 * @return A pointer to the appropriate exception instance
 */
extern ExceptionInstance_t* Exception_getInstance(void);
//--------------------------------------------------------------------------------------------
/** @ingroup exceptionHdl
 * @brief Handle Exception Error
 * Handler an exception error, e.g. when Throwing but you missed the Try-Catch statement
 *
 * @param    A pointer to the exception instance
 */
extern void handleExceptionError(ExceptionInstance_t * exInst);


//--------------------------------------------------------------------------------------------
/** @ingroup exceptionHdl
 * @brief Try for C
 * Abstraction of \e try using setjmp. The first letter is capitalized preventing confusion with \e try at
 * object-oriented languages.
 */
//TODO: The opening and closing curly brackets { } for catch and finally could also be replaced with an
//       "do{ ... }while(0);" statement when defining an END tag for both -> this could
//       prevent users from setting too much / too less brackets which would conflict with the catch/finally
//       ones. (the compiler can i.e. detect missing while(0) statements when there is a closing bracket too much)
//       furthermore it would help to ensure proper semicolon settings, see details here:
//       http://stackoverflow.com/questions/257418/do-while-0-what-is-it-good-for
#define Try                                                 \
do                                                          \
{                                                           \
    ExceptionInstance_t * exInst = Exception_getInstance(); \
    jmp_buf *pStackContentPrev, stackContent;               \
    pStackContentPrev = exInst->pStack;                     \
    exInst->pStack = &stackContent;                         \
    exInst->caught = 0;                                     \
    exInst->numbOfNestedTrys++;                             \
    if (setjmp(stackContent) == 0)                          \
    do                                                      \

//--------------------------------------------------------------------------------------------
/** @ingroup exceptionHdl
 * @brief Catch for C
 * Abstraction of \e catch referring to a thrown exception. The first letter is capitalized preventing confusion with \e catch
 * at object-oriented languages. Argument:
 * - \b e is an instance of type \ref Exception_t where the thrown exception is copied to
 * .
 */
#define Catch(e)                                                    \
    while(0);                                                       \
    else                                                            \
    {                                                               \
        exInst->caught = 1;                                         \
    }                                                               \
    exInst->pStack = pStackContentPrev;                             \
    exInst->numbOfNestedTrys--;                                     \
    memcpy(&e,(const void*)&(exInst->exception.current),sizeof(e)); \
} while(0);                                                         \
if (Exception_getInstance()->caught == 0) { }                       \
else if((Exception_getInstance()->caught = 0) > 0) {}               \
    else

//--------------------------------------------------------------------------------------------
/** @ingroup exceptionHdl
 * @brief Finally for C
 * Abstraction of \e finally referring to a thrown exception. The first letter is capitalized preventing confusion with \e finally
 * at object-oriented languages.\n
 * \e IMPORTANT: If you have nested Try-Throw-Catches and you don't want to rethrow the caught exception, you have to call finally!
 */
#define Finally                          \
    Exception_getInstance()->caught = 0; \
    if(1)


//--------------------------------------------------------------------------------------------
/** @ingroup exceptionHdl
 * @brief Throw for c
 * Abstraction of \e throw using longjmp. The first letter is capitalized preventing confusion with \e throw
 * at object-oriented languages. Arguments:\n
 * - \b exceptionId is one selected id out of \ref ExcpetionId_t
 * - \b args... are variable number of arguments (similar to printf) containing the exception message
 * .
 */
#define Throw(exceptId,...)                                                                \
do {                                                                                       \
    ExceptionInstance_t * exInst = Exception_getInstance();                                \
    exInst->exception.current.line = (__LINE__);                                           \
    exInst->exception.current.file = __FILE__;                                             \
    exInst->exception.current.function = __func__;                                         \
    exInst->exception.current.id = exceptId;                                               \
    snprintf((char*)(exInst->exception.current.msg), EXCEPTION_MSG_SIZE_MAX, __VA_ARGS__); \
    if(exInst->numbOfNestedTrys > 0) { longjmp(*exInst->pStack, 1); }                      \
    else { handleExceptionError(exInst); }                                                 \
} while(0)


#ifdef __cplusplus
}
#endif


#endif /* _EXCEPTION_H_ */

/************************************END OF FILE*********************************************/
