/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * Logging Interface HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : ILog.h
 * CREATION DATE : Mar 10, 2020
 * TARGET        : generic
 *
 *********************************************************************************************/

#ifndef ILOG_H_
#define ILOG_H_

#include <nmapp/types/nm-types.h>

/**
* \ingroup libnm
* \defgroup ILog Logging Interface
* Provides an interface for logging messages on a certain log level.
*/

#ifdef __cplusplus
extern "C" {
#endif


/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/**
 * Forward declaration for the interface reference.
 */
typedef struct __ILogPrivate * ILogRef;

/**
 * Log level definitions
 */
typedef enum __LogLevel
{
    LOG_LEVEL_EMCY  = 0,               //!< emergency - system is unusable
    LOG_LEVEL_ALERT = 1,               //!< alerting  - action must be taken immediately
    LOG_LEVEL_CRIT  = 2,               //!< critial   - critical condition
    LOG_LEVEL_ERR   = 3,               //!< error     - error condition
    LOG_LEVEL_WARN  = 4,               //!< warning   - warning condition
    LOG_LEVEL_NOTE  = 5,               //!< note      - notice = normal but significant condition
    LOG_LEVEL_INFO  = 6,               //!< info      - informational
    LOG_LEVEL_DBG   = 7,               //!< debug     - debug-level messages
    LOG_LEVEL_MAX
} LogLevel_t;



/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/** @ingroup ILog
 * @brief Log a Message
 * Send a log message to the implemented logger
 *
 * @param me        The reference to a logger instance (used to creating different log facilities)
 *                  Pass NULL = ILog (if there is one implemented, if not the function will assert)
 * @param logLevel  The log level (if the it is higher that the configured max log level the function will
 *                  return immediately -> resp. ignore the call and produce CPU load as low as possible)
 * @param msg       The log message format string (printf-like formating)
 */
void ILog_msg(ILogRef me, LogLevel_t logLevel, const char * msg, ...)
__attribute__ ((format (printf, 3, 4))); // this attribute enables compile-time checks of the format-string

//--------------------------------------------------------------------------------------------
/** @ingroup ILog
 * @brief Set Max Log Level
 * Set the max log level for this log facility (logs with higher log level will be ignored)
 *
 * @param me        The reference to a logger instance (used to creating different log facilities)
 *                  Pass NULL = ILog (if there is one implemented, if not the function will assert)
 * @param logLevel  The max log level
 */
void ILog_setMaxLogLevel(ILogRef me, LogLevel_t logLevel);

//--------------------------------------------------------------------------------------------
/** @ingroup ILog
 * @brief Get Max Log Level
 * Get the max log level used for this log facility
 * @param me        The reference to a logger instance (used to creating different log facilities)
 *                  Pass NULL = ILog (if there is one implemented, if not the function will assert)
 * @return          The max log level
 */
LogLevel_t ILog_getMaxLogLevel(ILogRef me);


#ifdef __cplusplus
}
#endif

#endif /* end of ILOG_H_ */
/************************************END OF FILE*********************************************/
