/*********************************************************************************************
 * (c) COPYRIGHT 2021 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * system-state MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : system-state.h
 * CREATION DATE : Jan 15, 2021
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/
#ifndef _SYSTEM_STATE_H_
#define _SYSTEM_STATE_H_


#include <nmapp/logging/ILog.h>
#include <nmapp/sys/IUnixInitSys.h>

/**
 * @defgroup system-state System-State
 * Provides the interface to the system state framework
 */
#ifdef __cplusplus
extern "C" {
#endif


/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/** SystemState object reference */
typedef struct __SystemStatePrivate * SystemStateRef;



/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/** @ingroup system-state
 * @brief Create System State Instance
 * Creates a system state instance
 *
 * @param logRef          A reference to the logger to use
 * @param initSysItfRef   A reference to the unix init-sys interface
 * @return The reference to the created instance
 */
SystemStateRef SystemState_create(ILogRef logRef, IUnixInitSysRef initSysItfRef);
//--------------------------------------------------------------------------------------------
/** @ingroup system-state
 * @brief Destroy System State Instance
 * Destroys a system state instance
 *
 * @param me     A reference to a system state instance
 */
void SystemState_destroy(SystemStateRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup system-state
 * @brief Initialize System State Instance
 * Initializes a system state instance
 *
 * @param me     A reference to a system state instance
 */
void SystemState_init(SystemStateRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup system-state
 * @brief Deinitialize System State Instance
 * Deinitializes a system state instance
 *
 * @param me     A reference to a system state instance
 */
void SystemState_deinit(SystemStateRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup system-state
 * @brief Get File Descriptor of the operating state
 * Returns the operating state file descriptor
 *
 * @param me     A reference to a system state instance
 * @return The file descriptor of the operating state monitor
 */
int SystemState_getOperatingStateFileDescriptor(SystemStateRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup system-state
 * @brief Operating State Change Callback (for Event Loop)
 * Callback for a change of the operating state
 *
 * @param me     A reference to a system state instance
 */
void SystemState_operatingStateChangeCallback(void *pCtx);
//--------------------------------------------------------------------------------------------
/** @ingroup system-state
 * @brief Is System Up
 * Returns whether the system is fully booted or not
 *
 * @param me     A reference to a system state instance
 * @return true=fully booted (even if an failure was detected), false otherwise
 */
bool SystemState_isSystemUp(SystemStateRef me);




#ifdef __cplusplus
}
#endif

#endif /* _SYSTEM_STATE_H_ */
