/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * List HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : list.h
 * CREATION DATE : Mar 30, 2020
 * TARGET        : generic
 *
 *********************************************************************************************/
#ifndef _LIST_H_
#define _LIST_H_

/** @ingroup libnmapp
 * @defgroup list List
 * Provides a container functionality using a list where loose list elements can be linked
 * to a list.
 */

#include <nmapp/types/nm-types.h>

/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/** Forward declaration for List object reference */
typedef struct ListPrivate* ListRef;

/** List element skeleton */
typedef struct ListElem {
    GenericTypeRef   content;         //!< Pointer to content of the element
    struct ListElem* prev;            //!< Pointer to previous element in the list
    struct ListElem* next;            //!< Pointer to next element in the list
} ListElem_t;


/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/

//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Create List
 * Creates a new empty list
 *
 * @return A reference to the created list
 */
ListRef List_create(void);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Destroy List
 * Destroys a list including all linked elements.
 *
 * @param me       A reference to the list
 */
void List_destroy(ListRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Add Element to List
 * Adds an element to a list
 *
 * @param listRef  A reference to the list
 * @param pElem    A reference to the element
 */
void List_addElement(ListRef listRef, ListElem_t* pElem);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Remove Element from List
 * Removes an element from a list. The element memory is only removed but not freed.
 *
 * @param listRef  A reference to the list
 * @param pElem    A reference to the element
 */
void List_removeElement(ListRef listRef, ListElem_t* pElem);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Remove And Destroy Element From List
 * Removes and destroys an element from a list
 *
 * @param listRef  A reference to the list
 * @param pElem    A reference to the element
 * @return The reference of the content of the removed element, the content memory is not freed
 */
GenericTypeRef List_removeAndDestroyElement(ListRef listRef, ListElem_t* pElem);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Remove And Destroy Element and its Content From List
 * Removes and destroys an element and its content from a list
 *
 * @param listRef  A reference to the list
 * @param pElem    A reference to the element
 */
void List_removeAndDestroyElementAndContent(ListRef listRef, ListElem_t* pElem);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Get First Element of List
 * Returns the first element in the list
 *
 * @param listRef       A reference to the list
 * @return The reference of the first element in the list
 */
ListElem_t* List_getFirstElement(ListRef listRef);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Get Last Element of List
 * Returns the last element in the list
 *
 * @param listRef       A reference to the list
 * @return The reference of the last element in the list
 */
ListElem_t* List_getLastElement(ListRef listRef);




//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Create List Element
 * Creates a new empty list element
 *
 * @return A reference to the created list element
 */
ListElem_t* List_createElement(void);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Create List Element
 * Creates a new empty list element
 *
 * @param pElem    A reference to the list element
 */
void List_destroyElement(ListElem_t* pElem);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Add Content to Element
 * Adds a content to an element
 *
 * @param pElem       A reference to the element
 * @param contentRef  A reference to the content
 */
void List_addContentToElement(ListElem_t* pElem, GenericTypeRef contentRef);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Remove Content from Element
 * Removes a content from an element
 *
 * @param pElem       A reference to the element
 * @return The reference of the content removed, the content memory was only removed but not freed
 */
GenericTypeRef List_removeContentFromElement(ListElem_t* pElem);




//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Create List Element Content
 * Creates a new empty list element content
 *
 * @param contentSize   The element content size in Bytes
 * @return A reference to the created list element content
 */
GenericTypeRef List_createElementContent(uint32_t contentSize);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Destroy List Element Content
 * Destroys a list element content
 *
 * @param pContent  A pointer to the list element content
 */
void List_destroyElementContent(GenericTypeRef pContent);




//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Create List Element And Its Content
 * Creates a new list element and its content
 *
 * @param contentSize   The element content size in Bytes
 * @return A reference to the created list element (content allocated)
 */
ListElem_t* List_createElementAndContent(uint32_t contentSize);
//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Destroy List Element And Its Content
 * Destroys a list element and its content
 *
 * @param pElem   A reference to the element
 */
void List_destroyElementAndContent(ListElem_t* pElem);




//--------------------------------------------------------------------------------------------
/** @ingroup list
 * @brief Destroy List
 * Removes and destroys all linked elements and its contents.
 *
 * @param listRef       A reference to the list
 */
void List_RemoveAndDestroyAllElementsAndContents(ListRef listRef);


#endif /* _LIST_H_ */

/************************************END OF FILE*********************************************/
