/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * IFileOperation HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : IFileOperation.h
 * CREATION DATE : Apr 2, 2020
 * TARGET        : generic
 *
 *********************************************************************************************/
#ifndef _FILE_IFILEOPERATION_H_
#define _FILE_IFILEOPERATION_H_

#include <sys/stat.h>
#include <nmapp/types/nm-types.h>

/** @ingroup libnmapp
 * @defgroup IFileOperation File Operation Interface
 * Provides an interface for file operations like move, copy, open, close, read, write ...
 */
#ifdef __cplusplus
extern "C" {
#endif


/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/** Reference to the created interface instance */
typedef struct __IFileOperationProtected * IFileOperationRef;

/** Forward declaration for the class which realizes this interface */
typedef void * IFileOperationImplRef;



/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Constructor of a File Operation Interface
 * Creates a file operation interface instance
 *
 * @param implRef           The reference to the realization of the interface
 * @param isFileOpen        Function pointer to the realized IFileOperation_isFileOpen function
 * @param openFile          Function pointer to the realized IFileOperation_openFile function
 * @param getFileName       Function pointer to the realized IFileoperation getFileName function
 * @param closeFile         Function pointer to the realized IFileOperation_closeFile function
 * @param writeToFile       Function pointer to the realized IFileOperation_writeToFile function
 * @param writeUpdateToFile Function pointer to the realized IFileOperation_writeUpdateToFile function
 * @param readFromFile      Function pointer to the realized IFileOperation_readFromFile function
 * @param readLineFromFile  Function pointer to the realized IFileOperation_readLineFromFile function
 * @param moveFile          Function pointer to the realized IFileOperation_moveFile function
 * @param deleteFilepath    Function pointer to the realized IFileOperation_deleteFilepath function
 * @param createDirectory   Function pointer to the realized IFileOperation_createDirectory function
 * @param isExist           Function pointer to the realized IFileOperation_isExist function
 * @param isDirectory       Function pointer to the realized IFileOperation_isDirectory function
 * @param getFiledescriptor Function pointer to the realized IFileOperation_getFiledescriptor function
 * @return The reference to the created instance
 */
IFileOperationRef IFileOperation_ctor(IFileOperationImplRef implRef,
                        bool (*isFileOpen)(IFileOperationImplRef implRef),
                        void (*openFile)(IFileOperationImplRef implRef, const char* fileName, const char* modes),
                        void (*closeFile)(IFileOperationImplRef implRef),
                        const char* (*getFileName)(IFileOperationImplRef implRef),
                        void (*writeToFile)(IFileOperationImplRef implRef, const char* data, uint32_t size),
                        void (*writeUpdateToFile)(IFileOperationImplRef implRef, bool withTruncate, uint32_t atOffset, const char* data, uint32_t size),
                        void (*readFromFile)(IFileOperationImplRef implRef, char* buffer, uint32_t size, uint32_t offset),
                        void (*readLineFromFile)(IFileOperationImplRef implRef, uint32_t lineNbr, char* buffer, uint32_t size),
                        void (*moveFile)(IFileOperationImplRef implRef, const char* srcFileName, const char* dstFileName),
                        void (*copyFile)(IFileOperationImplRef implRef, const char* srcFileName, const char* dstFileName),
                        void (*deleteFilepath)(IFileOperationImplRef implRef, const char* filePath),
                        void (*createDirectory)(IFileOperationImplRef implRef, const char* dirPath, uint32_t permissions),
                        bool (*isExist)(IFileOperationImplRef implRef, const char* fileName),
                        bool (*isDirectory)(IFileOperationImplRef implRef, const char* path),
                        int  (*getFiledescriptor)(IFileOperationImplRef implRef));
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Destructor of the File Operation interface instance
 * Destroys a file operation interface instance
 *
 * @param me           The reference to the interface instance
 */
void IFileOperation_destroy(IFileOperationRef me);

//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Checking if File is Open
 * Interface to check whether the file is open or not
 *
 * @param me           A reference to my instance
 * @return true = file is open, false otherwise
 */
bool IFileOperation_isFileOpen(IFileOperationRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Opening a File
 * Interface to open a file
 *
 * @param me           A reference to my instance
 * @param fileName     The pathname of the file
 */
void IFileOperation_open(IFileOperationRef me, const char* fileName, const char* modes);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Closing a File
 * Interface to close a file
 *
 * @param me           A reference to my instance
 */
void IFileOperation_close(IFileOperationRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Getting the File Name
 * Interface to get the file name
 *
 * @param me           A reference to my instance
 * @return The file name
 */
const char* IFileOperation_getFileName(IFileOperationRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Writing To a File
 * Interface to write to a file
 *
 * @param me           A reference to my instance
 * @param data         A pointer to the data
 * @param size         The size of the data
 */
void IFileOperation_writeToFile(IFileOperationRef me, const char* data, uint32_t size);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Writing an Update To a File
 * Interface to write an Update to a file, i.e. writes size bytes of data starting at position atOffset.
 *
 * @param me           A reference to my instance
 * @param withTruncate true to truncate the file after the size of the updating data, false to leave the rest
 * @param atOffset     The offset to start writing
 * @param data         A pointer to the data
 * @param size         The size of the data
 */
void IFileOperation_writeUpdateToFile(IFileOperationRef me, bool withTruncate,
                                      uint32_t atOffset, const char* data, uint32_t size);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Reading From a File
 * Interface to read from a file
 *
 * @param me           A reference to my instance
 * @param buffer       A pointer to the buffer where the data can be stored to
 * @param size         The size of data to read
 * @param offset       The offset from the beginning from where starting reading
 * @return see param buffer
 */
void IFileOperation_readFromFile(IFileOperationRef me, char* buffer, uint32_t size, uint32_t offset);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Reading a Line From a File
 * Interface to read a line from a file
 *
 * @param me           A reference to my instance
 * @param lineNbr      The number of the line to read (1st line means lineNbr = 1)
 * @param line         A pointer to the line buffer where the line can be stored to
 * @param size         The size of line buffer
 * @return see param line
 */
void IFileOperation_readLineFromFile(IFileOperationRef me, uint32_t lineNbr, char* line, uint32_t size);

//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Moving/Renaming File
 * Interface to move or rename a file
 *
 * @param me           A reference to my instance
 * @param srcFileName  The pathname of the source file
 * @param dstFileName  The pathname of the destination file
 */
void IFileOperation_move(IFileOperationRef me, const char* srcFileName, const char *dstFileName);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Copying a File
 * Interface to copy a file
 *
 * @param me           A reference to my instance
 * @param srcFileName  The pathname of the source file
 * @param dstFileName  The pathname of the destination file
 */
void IFileOperation_copy(IFileOperationRef me, const char* srcFileName, const char* dstFileName);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Deleting a Filepath
 * Interface to delete a filepath
 *
 * @param me           A reference to my instance
 * @param filePath     The path to delete
 */
void IFileOperation_delete(IFileOperationRef me, const char* filePath);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Creating a Directory
 * Interface to create a directory
 *
 * @param me           A reference to my instance
 * @param dirPath      The path of the directory to create
 * @param permissions  The permissions to set, e.g. (S_IRWXU|S_IRWXG|S_IRWXO) = 0777
 */
void IFileOperation_createDirectory(IFileOperationRef me, const char* dirPath, uint32_t permissions);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Checking if File Exists
 * Interface to check whether a file is existing or not
 *
 * @param me           A reference to my instance
 * @param fileName     The pathname of the file to check
 * @return true = file is existing, false otherwise
 */
bool IFileOperation_isExist(IFileOperationRef me, const char* fileName);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Checking if Path is a Directory
 * Interface to check whether a path is directory or not
 *
 * @param me           A reference to my instance
 * @param path         The path to check
 * @return true = path is a directory, false otherwise
 */
bool IFileOperation_isDirectory(IFileOperationRef me, const char* path);
//--------------------------------------------------------------------------------------------
/** @ingroup IFileOperation
 * @brief Interface for Getting the File Descriptor
 * Interface to get the file descriptor
 *
 * @param me           A reference to my instance
 * @return The file descriptor when file is open - throws an exception when not open or on error
 */
int IFileOperation_getFiledescriptor(IFileOperationRef me);


#ifdef __cplusplus
}
#endif

#endif /* _FILE_IFILEOPERATION_H_ */
