/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * System Logger Command Line HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : SysLogger-cmdline.h
 * CREATION DATE : Mar 26, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/
#ifndef _SYSLOGGER_CMDLINE_H_
#define _SYSLOGGER_CMDLINE_H_

#include <nmapp/logging/SysLoggerDefinitions.h>
#include <nmapp/logging/ILog.h>

/**
* \ingroup SysLogger
* \defgroup SysLoggerCmdLine System logger command line options
* Provides necessary functions for parsing logger definitions within the command line.
*/

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/** Getopt definitions for integrating into your getopt_long struct */
#define SYSLOGGER_LONGOPTS \
            {"disable-syslog", no_argument,       0, 0}, \
            {"enable-stdout",  no_argument,       0, 0}, \
            {"loglevel",       required_argument, 0, 0}


/** Command line option for a single logger definition */
typedef struct __SysLoggerOption
{
    char       loggerName[SYSLOG_MAX_LOGGER_NAME_LENGTH];  // Logger name
    LogLevel_t maxLogLevel;                                // Max log level for logger
} SysLoggerCmdLineOption_t;


/** Structure holding the logger definitions extracted from the command line */
typedef struct __SysLoggersCmdLine
{
    bool                     useSysLog;                              // General syslog enabled
    bool                     useStdout;                              // Standard out enabled
    char*                    appName;                                // The application name
    LogLevel_t               maxLogLevel;                            // General max log level
    SysLoggerCmdLineOption_t loggers[SYSLOG_MAX_NBR_OF_LOGGERS];     // Options for extracted loggers
} SysLoggersCmdLine_t;


/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Parse the command line for logger arguments
 * Parses the command line for logger definitions and returns a command
 * line structure holding the options for all extracted loggers.
 * @note the command line arguements are left untouched.
 *
 * @param argc   The amount of arguemnts
 * @param argv   The arguemnts list
 * @return The command line loggers
 */
SysLoggersCmdLine_t SysLogCmdLine_parse(int argc, char* argv[]);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Check if Logger Command line Option is Valid
 * check if a given command line option is a valid SysLogger command line option (to avoid
 * error handling in applications which are also parsing the same option arrays and don't
 * know the SysLogger option...)
 *
 * @param opt   the option to check
 * @return      true = if the option is a valid SysLogger option, false = otherwise
 */
bool SysLogCmdLine_isValidCommandLineOption(const char * opt);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Print Help
 * Print command line help text to stdout
 */
void SysLogCmdLine_printHelp(void);


#ifdef __cplusplus
}
#endif
#endif  //_SYSLOGGER_CMDLINE_H_
