/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * System Logger HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : SysLogger.h
 * CREATION DATE : Mar 24, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/
#ifndef _SYS_LOGGER__H_
#define _SYS_LOGGER__H_

#include <nmapp/types/nm-types.h>
#include <nmapp/logging/SysLogger-cmdline.h>
#include <nmapp/logging/ILog.h>

/**
* \ingroup libnm
* \defgroup SysLogger System Logger
* Implements the interface ILog and provides logger functionality, command line parsing
* helper functions and a set of definitions:
* - log facility
* - command line helper definition (getopt_long definition)
* - structures for logger options
*/

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/** SysLogger object reference */
typedef struct __SysLoggerPrivate * SysLoggerRef;



/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Create System Logger
 * Create an instance of the SysLogger adapter (using syslogd) - Singleton.
 *
 * @param appName      The application name
 * @param maxLogLevel  The max log level for the system logger
 * @return The created instance
 */
SysLoggerRef SysLogger_create(const char* appName, LogLevel_t maxLogLevel);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Destroy System Logger
 * Destructor of the SysLogger which is freeing any allocated memory.
 *
 * @param me    The reference to the SysLogger object
 */
void SysLogger_destroy(SysLoggerRef me);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Add Command Line Loggers
 * Add the loggers configured over the command line
 *
 * @param cmdLineCfgRef A reference to the command line parsed configuration (see SysLogCmdLine_parse)
 */
void SysLogger_addCommandLineLoggers(SysLoggersCmdLine_t* cmdLineCfgRef);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Get the logger interface
 * Get the log interface (ILog) for this SysLogegr instance.
 * @param me    The reference to the SysLogger object
 * @return
 */
ILogRef SysLogger_getLoggerItf(SysLoggerRef me);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Get the logger interface of a logger instance
 * Get the ILog interface of an already existing logger instance or  optionally
 * create it if missing (parameter addIfMissing = true). This is a wrapper function
 * around SysLogger_addLogger for convenience (mostly the detailed configuration
 * of a logger instance is not needed and it can simply inherit everything except
 * the name from the global syslogger)
 *
 * @param name          The logger name to use (will be part of the LOG message
 *                      and is therefore usable like a tag/filter criteria)
 * @param addIfMissing  true ->
 *                          The logger will be created with default configuration
 *                          of the global sysLogger. The default parameters are
 *                          logFacility, maxLogLevel, useSyslog and useStdOut.
 *                      false ->
 *                          NULL is returned if the logger does not exist
 * @return A reference to the logger instance interface OR NULL if not existing
 */
ILogRef SysLogger_getLoggerInstance(const char * name, bool addIfMissing);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Add a new Logger
 * Create a new logger with a given configuration and get it's ILog interface
 * @param name          The logger name to use
 *                      -> Will be part of the LOG message and is therefore usable like a tag/filter criteria
 * @param logFacility   The syslog facility for this logger
 *                      Attention: heavy use of different log facilities in a single application will
 *                      increase the CPU load since the facility needs to be (re)opened on facility changes.
 *                      Worst case would be toggling log messages with different facilities.
 * @param maxLogLevel   The max log level for this specific logger (there can be different configurations
 *                      for each logger within one appilcation
 * @param useSyslog     true=use the syslog as log output, false=do not route the messages to syslog
 * @param useStdout     true=print log messages on stdout use the, false=do not print log messages on stdout
 * @return              The reference of the Ilog interface for the new logger
 */
ILogRef SysLogger_addLogger(const char * name, SysLogFacility_t logFacility, LogLevel_t maxLogLevel, bool useSyslog, bool useStdout);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Set the Max Log Level
 * Set the max log level for the SysLogger (not the added loggers, this is handled in the ILog interface!)
 *
 * @param me           The reference to the SysLogger object
 * @param maxLoglevel  The configured log level
 */
void SysLogger_setMaxLogLevel(SysLoggerRef me, LogLevel_t maxLogLevel);

//--------------------------------------------------------------------------------------------
/** @ingroup SysLogger
 * @brief Get the Max Log Level
 * Read the configured max log level for the SysLogger (not the added loggers, this is handled in the ILog interface!)
 *
 * @param me    The reference to the SysLogger object
 * @return      The configured log level
 */
LogLevel_t SysLogger_getMaxLogLevel(SysLoggerRef me);


#ifdef __cplusplus
}
#endif

#endif
/************************************END OF FILE*********************************************/

