/*********************************************************************************************
 * (c) COPYRIGHT 2021 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * IUnixInitSys HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : IUnixInitSys.h
 * CREATION DATE : Jan 20, 2021
 * TARGET        : generic
 *
 *********************************************************************************************/
#ifndef _SYS_IUNIXINITSYS_H_
#define _SYS_IUNIXINITSYS_H_

/** @ingroup libnmapp
 * @defgroup IUnixInitSys Unit Init-Sys Interface
 * Provides an interface for the unit init-system operations
 */
#ifdef __cplusplus
extern "C" {
#endif


/*********************************************************************************************
 *        GLOBAL DECLARATIONS
 *********************************************************************************************/

/** Reference to the created interface instance */
typedef struct __IUnixInitSysProtected * IUnixInitSysRef;

/** Forward declaration for the class which realizes this interface */
typedef void * IUnixInitSysImplRef;

/** Forward declaration for dbus signal callback
 * void pointer = callback context
 * const char pointers = strings to dbus svc/destination, path, interface, signal and item
 */
typedef void (*UnixInitSysDBusSignalItemCallback)(void*, const char*, const char*, const char*, const char*, const char*);


/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Constructor of a File Operation Interface
 * Creates a file operation interface instance
 *
 * @param implRef                      The reference to the realization of the interface
 * @param logMessage                   Function pointer to the realized IUnixInitSys_logMessage function
 * @param initDBus                     Function pointer to the realized IUnixInitSys_initDBus function
 * @param deinitDBus                   Function pointer to the realized IUnixInitSys_deinitDBus function
 * @param getEventFd                   Function pointer to the realized IUnixInitSys_getEventFileDescriptor function
 * @param registerDBusSignalCallback   Function pointer to the realized IUnixInitSys_registerDBusSignalCallback function
 * @param activateDBusListener         Function pointer to the realized IUnixInitSys_activateDBusListerner function
 * @param processDBusEvents            Function pointer to the realized IUnixInitSys_processDBusEvents function
 * @param registerDBusSignalMatch      Function pointer to the realized IUnixInitSys_registerDBusSignalMatch function
 * @param unregisterDBusSignalMatch    Function pointer to the realized IUnixInitSys_unregisterDBusSignalMatch function
 * @param getOperatingState            Function pointer to the realized IUnixInitSys_getOperatingState function
 * @return The reference to the created instance
 */
IUnixInitSysRef IUnixInitSys_ctor(IUnixInitSysImplRef implRef,
                                  void  (*logMessage)(IUnixInitSysImplRef implRef, int32_t logPriority, int32_t facility, const char* loggerName, const char* msg),
                                  void  (*initDBus)(IUnixInitSysImplRef implRef),
                                  void  (*deinitDBus)(IUnixInitSysImplRef implRef),
                                  int   (*getEventFd)(IUnixInitSysImplRef implRef),
                                  void  (*registerDBusSignalCallback)(IUnixInitSysImplRef implRef, UnixInitSysDBusSignalItemCallback cbFn, void* pCbCtx),
                                  void  (*activateDBusListener)(IUnixInitSysImplRef implRef),
                                  void  (*processDBusEvents)(IUnixInitSysImplRef implRef),
                                  void* (*registerDBusSignalMatch)(IUnixInitSysImplRef implRef, const char* dbusService, const char* dbusObjPath,
                                                                   const char* dbusItfName, const char* dbusSignalName),
                                  void  (*unregisterDBusSignalMatch)(IUnixInitSysImplRef implRef, void* signalMatchRef),
                                  const char* (*getOperatingState)(IUnixInitSysImplRef me));
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Destructor of the File Operation interface instance
 * Destroys a file operation interface instance
 *
 * @param me           The reference to the interface instance
 */
void IUnixInitSys_destroy(IUnixInitSysRef me);


//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for Logging a Message
 * Interface for logging a message to the system log, i.e. for systemd into the journal.
 *
 * @param me             A reference to my instance
 * @param logPriority    The log level/priority
 * @param facility       The number of the facility
 * @param loggerName     The name of the logger
 * @param msg            The message to log
 */
void IUnixInitSys_logMessage(IUnixInitSysRef me, int32_t logPriority, int32_t facility, const char* loggerName, const char* msg);

//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for initialize DBus
 * Interface to init of the dbus.
 * Monitoring the init-sys operating state for states: starting, running, up but failed
 *
 * @param me           A reference to my instance
 */
void IUnixInitSys_initDBus(IUnixInitSysRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for de-initialize DBus
 * Interface to de-init of the dbus.
 *
 * @param me           A reference to my instance
 */
void IUnixInitSys_deinitDBus(IUnixInitSysRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for Getting the Event File Descriptor
 * Interface to get event file descriptor
 *
 * NOTE: If you are using evt-loop, do not use this file descriptor getter
 * beside the way it is used in evt-loop.
 *
 * @param me           A reference to my instance
 * @return The file descriptor of the event
 */
int IUnixInitSys_getEventFileDescriptor(IUnixInitSysRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for Activating the DBus Listener
 * Interface to activate the dbus listener (sd_event_prepare)
 *
 * @param me           A reference to my instance
 */
void IUnixInitSys_activateDBusListener(IUnixInitSysRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for registering DBUS signal callback
 * Interface to register a dbus signal callback. Events of dbus are triggered to one single
 * callback. Due to the encapsulation of systemd this callback cannot be pusblished since
 * there are data types necessary of systemd. Hence a callback can be registered by using this
 * function. On a dbus signal match, the internal callback will call this registered callback.
 *
 * @param me           A reference to my instance
 * @param cbFn         A function pointer to the external signal item callback
 * @param pCbCtx       A pointer to the callback context
 */
void IUnixInitSys_registerDBusSignalCallback(IUnixInitSysRef me, UnixInitSysDBusSignalItemCallback cbFn, void* pCbCtx);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for processing DBus events
 * Interface to process events on the dbus. Needs to be called after getting an event to get
 * triggered at the next events.
 *
 * @param me           A reference to my instance
 */
void IUnixInitSys_processDBusEvents(IUnixInitSysRef me);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Register DBus Signal Match
 * Registering a dbus signal match
 *
 * @param me             A reference to my instance
 * @param dbusService    The dbus service (string), e.g. "org.freedesktop.systemd1"
 * @param dbusObjPath    The dbus object path, e.g. "/org/freedesktop/systemd1"
 * @param dbusItfName    The dbus interface name, e.g. "org.freedesktop.systemd1.Manager"
 * @param dbusSignalName The dbus signal name, e.g. "UnitNew"
 * @return A reference to the dbus signal match (casted sd_bus_slot pointer)
 */
void* IUnixInitSys_registerDBusSignalMatch(IUnixInitSysRef me, const char* dbusService, const char* dbusObjPath,
                                          const char* dbusItfName, const char* dbusSignalName);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Unregister DBus Signal Match
 * Unregistering a dbus signal match
 *
 * @param me             A reference to my instance
 * @param signalMatchRef A reference to the dbus signal match (returned value of IUnixInitSys_registerDBusSignalMatch)
 */
void IUnixInitSys_unregisterDBusSignalMatch(IUnixInitSysRef me, void* signalMatchRef);
//--------------------------------------------------------------------------------------------
/** @ingroup IUnixInitSys
 * @brief Interface for Getting the Operating State
 * Interface to get the operating state
 *
 * @param me           A reference to my instance
 * @return The read operating state string
 */
const char* IUnixInitSys_getOperatingState(IUnixInitSysRef me);


#ifdef __cplusplus
}
#endif

#endif /* _SYS_IUNIXINITSYS_H_ */
