/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * List
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : list.c
 * CREATION DATE : Mar 30, 2020
 * TARGET        : generic
 *
 *********************************************************************************************/

#include <stdlib.h>                    // for calloc
#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>

#include "list-private.h"

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/


//--------------------------------------------------------------------------------------------
bool isListEmpty(ListRef lst)
{
    return (lst->pLastElem == NULL);
}
//--------------------------------------------------------------------------------------------
bool isElementLinked(ListElem_t* pElement)
{
    return ((pElement->prev != NULL) || (pElement->next != NULL));
}
//--------------------------------------------------------------------------------------------
bool isLastElement(ListElem_t* pElement)
{
    return (pElement->next == NULL);
}
//--------------------------------------------------------------------------------------------
bool isFirstElement(ListElem_t* pElement)
{
    return (pElement->prev == NULL);
}


/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
ListRef List_create(void)
{
    ListRef newList = (ListRef)calloc(1, sizeof(ListPrivate_t));
    if(newList == NULL)
    {
        Throw(EX_CREATE, "Creating new list");
    }
    return newList;
}
//--------------------------------------------------------------------------------------------
void List_destroy(ListRef listRef)
{
    if (listRef != NULL)
    {
        List_RemoveAndDestroyAllElementsAndContents(listRef);
        free(listRef);
    }
}


//--------------------------------------------------------------------------------------------
void List_addElement(ListRef listRef, ListElem_t* pElem)
{
    assert(listRef != NULL);
    assert(pElem != NULL);

    if (isListEmpty(listRef))
    {
        pElem->prev = NULL;
        pElem->next = NULL;
        listRef->pFirstElem = pElem;
    }
    else
    {
        listRef->pLastElem->next = pElem;
        pElem->prev = listRef->pLastElem;
    }
    listRef->pLastElem = pElem;
}
//--------------------------------------------------------------------------------------------
void List_removeElement(ListRef listRef, ListElem_t* pElem)
{
    assert(listRef != NULL);
    assert(pElem != NULL);

    if(isElementLinked(pElem))
    {
        if(isLastElement(pElem))
        {
            listRef->pLastElem = pElem->prev;
            pElem->prev->next  = NULL;
            return;
        }
        if(isFirstElement(pElem))
        {
            listRef->pFirstElem = pElem->next;
            pElem->next->prev   = NULL;
            return;
        }

        // we are in the middle of the list
        pElem->prev->next = pElem->next;
        pElem->next->prev = pElem->prev;
        return;
    }

    listRef->pFirstElem = NULL;
    listRef->pLastElem  = NULL;
}
//--------------------------------------------------------------------------------------------
GenericTypeRef List_removeAndDestroyElement(ListRef listRef, ListElem_t* pElem)
{
    assert(listRef != NULL);
    assert(pElem != NULL);

    GenericTypeRef contentRef = pElem->content;
    List_removeElement(listRef, pElem);
    List_destroyElement(pElem);
    return contentRef;
}
//--------------------------------------------------------------------------------------------
void List_removeAndDestroyElementAndContent(ListRef listRef, ListElem_t* pElem)
{
    List_removeElement(listRef, pElem);
    List_destroyElementAndContent(pElem);
}
//--------------------------------------------------------------------------------------------
ListElem_t* List_getFirstElement(ListRef listRef)
{
    assert(listRef != NULL);
    return listRef->pFirstElem;
}
//--------------------------------------------------------------------------------------------
ListElem_t* List_getLastElement(ListRef listRef)
{
    assert(listRef != NULL);
    return listRef->pLastElem;
}




//--------------------------------------------------------------------------------------------
ListElem_t* List_createElement(void)
{
    ListElem_t* newElement = (ListElem_t*)calloc(1, sizeof(ListElem_t));
    if (newElement == NULL)
    {
        Throw(EX_CREATE, "Creating new list element");
    }
    newElement->content = NULL;
    newElement->prev    = NULL;
    newElement->next    = NULL;
    return newElement;
}
//--------------------------------------------------------------------------------------------
void List_destroyElement(ListElem_t* pElem)
{
    if (pElem != NULL)
    {
        free(pElem);
    }
}
//--------------------------------------------------------------------------------------------
void List_addContentToElement(ListElem_t* pElem, GenericTypeRef contentRef)
{
    assert(pElem != NULL);
    if (pElem->content != NULL)
    {
        Throw(EX_ADD, "Adding content to list element");
    }
    pElem->content = contentRef;
}
//--------------------------------------------------------------------------------------------
GenericTypeRef List_removeContentFromElement(ListElem_t* pElem)
{
    GenericTypeRef contentRef = pElem->content;
    pElem->content = NULL;
    return contentRef;
}




//--------------------------------------------------------------------------------------------
GenericTypeRef List_createElementContent(uint32_t contentSize)
{
    GenericTypeRef newContent = calloc(1, contentSize);
    if (newContent == NULL)
    {
        Throw(EX_CREATE, "Creating new list element content");
    }
    return newContent;
}
//--------------------------------------------------------------------------------------------
void List_destroyElementContent(GenericTypeRef pContent)
{
    if (pContent != NULL)
    {
        free(pContent);
    }
}




//--------------------------------------------------------------------------------------------
ListElem_t* List_createElementAndContent(uint32_t contentSize)
{
    ListElem_t* pNewElement = List_createElement();
    pNewElement->content = List_createElementContent(contentSize);
    return pNewElement;
}
//--------------------------------------------------------------------------------------------
void List_destroyElementAndContent(ListElem_t* pElem)
{
    if (pElem != NULL)
    {
        List_destroyElementContent(pElem->content);
        pElem->content = NULL;
    }
    List_destroyElement(pElem);
}



//--------------------------------------------------------------------------------------------
void List_RemoveAndDestroyAllElementsAndContents(ListRef listRef)
{
    assert(listRef != NULL);

    ListElem_t* pElem = listRef->pFirstElem;
    while(pElem != NULL)
    {
        List_removeAndDestroyElementAndContent(listRef, pElem);
        pElem = listRef->pFirstElem;
    }
}




//============================================================================================
//    FOR TESTS ONLY:
//---------------------------------

//--------------------------------------------------------------------------------------------
void List_print(ListRef listRef)
{
    assert(listRef != NULL);
    ListElem_t* pElem = listRef->pFirstElem;
    uint32_t    nbr = 0;

    while(pElem != NULL)
    {
        nbr++;
        printf("%d:"
               "  addr = %p\n"
               "  next = %p\n"
               "  prev = %p\n"
               "----------------------\n", (int)nbr, (void*)pElem, (void*)pElem->next, (void*)pElem->prev);
        pElem = pElem->next;
    }
    fflush(stdout);

}
//--------------------------------------------------------------------------------------------
void List_setupNotEmptyList(ListRef listRef)
{
    assert(listRef != NULL);
    ListElem_t* elemRef = List_createElement();
    List_addElement(listRef, elemRef);
}


/************************************END OF FILE*********************************************/
