/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * Event Loop Private HEADER
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : evt-loop-private.h
 * CREATION DATE : Mar 30, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/
#ifndef _EVT_LOOP_PRIVATE_H_
#define _EVT_LOOP_PRIVATE_H_


#include <event2/event.h>
//#include <nmapp/container/list.h>
#include <nmapp/event/evt-loop.h>


#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************
 *        PRIVATE DECLARATIONS
 *********************************************************************************************/

// EventTypes:
typedef enum
{
    EVTLOOP_NONE         = 0,          // no event
    EVTLOOP_SIGNAL       = 1,          // signal event like SIGTERM, ...
    EVTLOOP_TIMER        = 2,          // timer event
    EVTLOOP_EPOLL        = 3,          // epoll event, i.e. edge triggered
    EVTLOOP_REGULAR_FILE = 4,          // regular file change, i.e. ready to read from / ready to write to
    EVTLOOP_DBUS         = 5           // dbus event
} EvtLoopEventType_t;

// dbus definitions:
typedef struct __EvtLoopDBusSignalId
{
    const char* destination;
    const char* objPath;
    const char* itf;
    const char* signal;
} EvtLoopDBusSignalId_t;

typedef struct __EvtLoopDBusEvtPrivate
{
    const char*           evtItem;
    void*                 busSlot;
    EvtLoopDBusSignalId_t busId;
    IUnixInitSysRef       initSysRef;
} EvtLoopDBusEvt_t;


// Event content that will be added to the event loop:
typedef struct __EvtLoopEventPrivate
{
    ILogRef              logRef;
    EvtLoopEventType_t   evtType;
    int                  fd;
    const char*          filePath;
    EvtLoopDBusEvt_t     dbusEvt;
    GenericTypeRef       contextRef;
    EvtLoopEventCallback evtCallbackFn;
    struct event*        evtRef;
} EvtLoopEvent_t;


// Timer event content:
typedef struct __EvtLoopTimerPrivate
{
    EvtLoopEvent_t* evtRef;
    struct timeval  timeout;
    bool            isPeriodic;
    bool            isRunning;
} EvtLoopTimer_t;


// Hash Table Entry Structures
typedef struct __EvtListEntry
{
    int             fd;
    EvtLoopEventRef evtRef;
    UT_hash_handle  hh;
} EvtListEntry_t;



// Event loop instance content:
typedef struct __EvtLoopPrivate
{
    ILogRef             logRef;
    struct event_base*  evtBaseRef;
    bool                hasExitSignals;

    int                 epollFd;             // edge triggered events
    struct event*       epollEvtRef;
    uint32_t            epollEvtCnt;

    int                 inotifyFd;           // file system monitoring events
    struct event*       inotifyEvtRef;
    uint32_t            inotifyEvtCnt;

    int                 dbusFd;              // dbus events
    IUnixInitSysRef     initSysItfRef;
    struct event*       dbusEvtRef;
    uint32_t            dbusEvtCnt;

    EvtListEntry_t*     signalListRef;
    EvtListEntry_t*     epollListRef;
    EvtListEntry_t*     inotifyListRef;
    EvtListEntry_t*     dbusListRef;
} EvtLoop_t;


/*********************************************************************************************
 *        PRIVATE FUNCTIONS
 *********************************************************************************************/

extern void signalEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx);
extern void epollEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx);
extern void filesystemChangedEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx);
extern void timerTimoutCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx);
extern void dbusEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx);
extern void dbusSignalMatchCallbackFn(void* ctx, const char* dbusSvc, const char* dbusPath, const char* dbusItf,
                                      const char* dbusSignal, const char* dbusSignalItem);

#ifdef __cplusplus
}
#endif

#endif /* SRC_EVENT_EVT_LOOP_PRIVATE_H_ */
