/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * Event Loop MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : evt-loop.c
 * CREATION DATE : Mar 30, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <signal.h>       // for signal numbers
#include <errno.h>        // for errno
#include <unistd.h>       // for read, ...
#include <sys/epoll.h>    // for epoll structures and definitions
#include <sys/inotify.h>  // for inotify structures and definitions --> filesystem monitoring

#include "evt-loop-private.h"

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/
#define INOTIFY_EVENT_SIZE   (sizeof(struct inotify_event))
#define INOTIFY_BUFFER_SIZE  (1024 * (INOTIFY_EVENT_SIZE + 16))

const int DBUS_FD_BASE_NBR = 90000;

//--------------------------------------------------------------------------------------------
static void terminateApplication(void *pCtx)
{
    EvtLoopRef me = (EvtLoopRef)pCtx;
    if(me == NULL)
    {
        ILog_msg(me->logRef, LOG_LEVEL_ERR, "TerminateApplication callback without context called");
        return;
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "Terminating application");
    event_base_loopexit(me->evtBaseRef, 0);
}


//--------------------------------------------------------------------------------------------
void createListEntryAndAddEventToList(EvtListEntry_t** pListRef, EvtLoopEventRef evtRef)
{
    assert(evtRef != NULL);
    EvtListEntry_t* entryRef = calloc(1, sizeof(EvtListEntry_t));
    if(entryRef == NULL)
    {
        Throw(EX_CREATE, "Cannot allocate hash table entry for fd=%d", evtRef->fd);
    }
    entryRef->fd     = evtRef->fd;
    entryRef->evtRef = evtRef;
    HASH_ADD(hh, *pListRef, fd, (int)sizeof(evtRef->fd), entryRef);
}
//--------------------------------------------------------------------------------------------
bool isDBusSignalMatchListEntry(EvtListEntry_t* pEntry, const char* dest, const char* path, const char* itf,
                                const char* sig, const char* item)
{
    assert(path != NULL);
    assert(itf != NULL);
    assert(sig != NULL);

    // NOTE:
    // dest and item can be NULL depending on its source and when you are testing if the signal
    // matches an entry.

    bool isItem = (item == NULL) ? true : (strncmp(item, pEntry->evtRef->dbusEvt.evtItem, strlen(item)) == 0);
    bool isDest = (dest == NULL) ? true : (strncmp(dest, pEntry->evtRef->dbusEvt.busId.destination, strlen(dest)) == 0);
    if( isDest && isItem &&
        (strncmp(path, pEntry->evtRef->dbusEvt.busId.objPath, strlen(path)) == 0) &&
        (strncmp(itf,  pEntry->evtRef->dbusEvt.busId.itf, strlen(itf)) == 0) &&
        (strncmp(sig,  pEntry->evtRef->dbusEvt.busId.signal, strlen(sig)) == 0) )
    {
        return true;
    }
    return false;
}
//--------------------------------------------------------------------------------------------
EvtListEntry_t* iterateThroughListAndGetMatchedEntry(EvtListEntry_t** pListRef, const char* dest, const char* path,
                                                     const char* itf, const char* sig, const char* item)
{
    EvtListEntry_t* pStore = NULL;
    EvtListEntry_t* pEntry = NULL;
    HASH_ITER(hh, *pListRef, pEntry, pStore)
    {
        if(isDBusSignalMatchListEntry(pEntry, dest, path, itf, sig, item))
        {
            return pEntry;
        }
    }
    return NULL;
}
//--------------------------------------------------------------------------------------------
void createDBusListEntryAndAddEventToList(EvtListEntry_t** pListRef, EvtLoopEventRef evtRef)
{
    EvtListEntry_t* pEntry = iterateThroughListAndGetMatchedEntry(pListRef,
                                                                  evtRef->dbusEvt.busId.destination, evtRef->dbusEvt.busId.objPath,
                                                                  evtRef->dbusEvt.busId.itf, evtRef->dbusEvt.busId.signal, NULL);

    evtRef->dbusEvt.busSlot = (pEntry != NULL) ?
                                  pEntry->evtRef->dbusEvt.busSlot :
                                  IUnixInitSys_registerDBusSignalMatch(evtRef->dbusEvt.initSysRef,
                                                                       evtRef->dbusEvt.busId.destination,
                                                                       evtRef->dbusEvt.busId.objPath,
                                                                       evtRef->dbusEvt.busId.itf,
                                                                       evtRef->dbusEvt.busId.signal);
    createListEntryAndAddEventToList(pListRef, evtRef);
}


//--------------------------------------------------------------------------------------------
void signalEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx)
{
    assert(ctx != NULL);
    EvtLoopRef me = (EvtLoopRef)ctx;
    EvtListEntry_t* entryRef = NULL;
    int sigNbr = (int)fd;

    HASH_FIND(hh, me->signalListRef, &sigNbr, (int)sizeof(sigNbr), entryRef);
    if(entryRef == NULL)
    {
        ILog_msg(me->logRef, LOG_LEVEL_WARN, "Received signal callback of unregistered signal %d", sigNbr);
        return;
    }
    assert(entryRef->evtRef->fd == sigNbr);
    if(entryRef->evtRef->evtCallbackFn != NULL)
    {
        entryRef->evtRef->evtCallbackFn(entryRef->evtRef->contextRef);
        return;
    }
}
//--------------------------------------------------------------------------------------------
void epollEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx)
{
    assert(ctx != NULL);
    EvtLoopRef me = (EvtLoopRef)ctx;

    int waitForever = -1;
    struct epoll_event events[me->epollEvtCnt];
    int nbrOfFds = epoll_wait(me->epollFd, events, (int)me->epollEvtCnt, waitForever);
    if (nbrOfFds == -1)
    {
        Throw(EX_READ, "epoll_wait error %d: %s", errno, strerror(errno));
    }

    int  i = 0;
    bool isEvtHandled;
    EvtListEntry_t* entryRef = NULL;
    for(i = 0; i < nbrOfFds; i++)
    {
        isEvtHandled = false;
        HASH_FIND(hh, me->epollListRef, &events[i].data.fd, (int)sizeof(events[i].data.fd), entryRef);
        if(entryRef != NULL)
        {
            assert(entryRef->evtRef->fd == events[i].data.fd);
            isEvtHandled = true;
            if(entryRef->evtRef->evtCallbackFn != NULL)
            {
                ILog_msg(me->logRef, LOG_LEVEL_DBG, "Calling epoll callback for file with fd=%d", entryRef->evtRef->fd);
                entryRef->evtRef->evtCallbackFn(entryRef->evtRef->contextRef);
                break;
            }
        }
        if(!isEvtHandled)
        {
            ILog_msg(me->logRef, LOG_LEVEL_WARN, "Received epoll callback of unregistered file descriptor %d", events[i].data.fd);
        }
    }// end looping through number of changed file descriptors
}
//--------------------------------------------------------------------------------------------
void filesystemChangedEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx)
{
    assert(ctx != NULL);
    EvtLoopRef me = (EvtLoopRef)ctx;
    char inotifyBuffer[INOTIFY_BUFFER_SIZE];

    ssize_t bytesRead = read(me->inotifyFd, inotifyBuffer, INOTIFY_BUFFER_SIZE);
    if(bytesRead < 0)
    {
        Throw(EX_READ, "inotify read error %d: %s", errno, strerror(errno));
    }
    ssize_t i = 0;
    bool    isEvtHandled;
    EvtListEntry_t* entryRef = NULL;
    struct inotify_event *inotfiyEventRef = NULL;

    while(i < bytesRead)
    {
        inotfiyEventRef = (struct inotify_event*)&inotifyBuffer[i];
        isEvtHandled = false;
        HASH_FIND(hh, me->inotifyListRef, &inotfiyEventRef->wd, (int)sizeof(inotfiyEventRef->wd), entryRef);
        if(entryRef != NULL)
        {
            assert(entryRef->evtRef->fd == inotfiyEventRef->wd);
            isEvtHandled = true;
            if(entryRef->evtRef->evtCallbackFn != NULL)
            {
                ILog_msg(me->logRef, LOG_LEVEL_DBG, "Test: File %s was modified", inotfiyEventRef->name);
                ILog_msg(me->logRef, LOG_LEVEL_DBG, "Calling inotify callback for file %s", entryRef->evtRef->filePath);
                entryRef->evtRef->evtCallbackFn(entryRef->evtRef->contextRef);
                return;
            }
        }
        i += (ssize_t)(INOTIFY_EVENT_SIZE + inotfiyEventRef->len);
        if(!isEvtHandled)
        {
            ILog_msg(me->logRef, LOG_LEVEL_WARN, "Received inotify callback of unregistered file %s", inotfiyEventRef->name);
        }
    }// end looping through number of read bytes
}
//--------------------------------------------------------------------------------------------
void timerTimoutCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx)
{
    assert(ctx != NULL);
    EvtLoopTimerRef me = (EvtLoopTimerRef)ctx;
    if(evtFlags != EV_TIMEOUT)
    {
        Throw(EX_INVALID_ARG, "Timeout callback called that is not a timeout event");
    }

    if( (me->evtRef != NULL) && (me->evtRef->evtCallbackFn != NULL) )
    {
        ILog_msg(me->evtRef->logRef, LOG_LEVEL_DBG, "Calling registered user timer callback (timeout=%us,%uus)...",
                                                     (uint32_t)me->timeout.tv_sec,
                                                     (uint32_t)me->timeout.tv_usec);
        me->evtRef->evtCallbackFn(me->evtRef->contextRef);
    }

    if( (me->isPeriodic) && (me->isRunning) )
    {
        EvtLoop_startTimer(me);
        return;
    }
    me->isRunning = false;
}
//--------------------------------------------------------------------------------------------
void dbusEventCallbackFn(evutil_socket_t fd, short evtFlags, void *ctx)
{
    assert(ctx != NULL);
    EvtLoopRef me = (EvtLoopRef)ctx;

    // dbus events are processed by the unix init system module. During this
    //  processing mechanism the registered event triggers the callback then.
    IUnixInitSys_processDBusEvents(me->initSysItfRef);
}
//--------------------------------------------------------------------------------------------
void dbusSignalMatchCallbackFn(void* ctx, const char* dbusDest, const char* dbusPath, const char* dbusItf,
                               const char* dbusSignal, const char* dbusSignalItem)
{
    assert(ctx != NULL);
    assert(dbusSignal != NULL);
    assert(dbusPath != NULL);
    assert(dbusItf != NULL);
    EvtLoopRef me = (EvtLoopRef)ctx;

    ILog_msg(me->logRef, LOG_LEVEL_DBG, "dbus signal match callback: svc=%s, path=%s, itf=%s, signal=%s, item=%s",
                                        dbusDest, dbusPath, dbusItf, dbusSignal, dbusSignalItem);

    EvtListEntry_t* pEntry = iterateThroughListAndGetMatchedEntry(&me->dbusListRef, dbusDest, dbusPath, dbusItf,
                                                                  dbusSignal, dbusSignalItem);
    if(pEntry == NULL)
    {
        ILog_msg(me->logRef, LOG_LEVEL_DBG,"dbus signal callback called without a registration: name=%s, item=%s",
                                            dbusSignal, dbusSignalItem);
        return;
    }

    if(pEntry->evtRef->evtCallbackFn != NULL)
    {
        ILog_msg(me->logRef, LOG_LEVEL_DBG, "calling dbus signal callback for signal/item=%s/%s",
                                            pEntry->evtRef->dbusEvt.busId.signal, pEntry->evtRef->dbusEvt.evtItem);
        pEntry->evtRef->evtCallbackFn(pEntry->evtRef->contextRef);
        return;
    }

    ILog_msg(me->logRef, LOG_LEVEL_DBG, "no callback function for dbus signal/item=%s/%s",
                                        pEntry->evtRef->dbusEvt.busId.signal, pEntry->evtRef->dbusEvt.evtItem);
}



//--------------------------------------------------------------------------------------------
void removeSignalElement(EvtLoopRef me, EvtListEntry_t* entryRef)
{
    assert(me != NULL);
    assert(entryRef != NULL);

    EvtLoopEventRef sigEvtRef = entryRef->evtRef;
    if(sigEvtRef->evtRef != NULL)
    {
        event_del(sigEvtRef->evtRef);
        event_free(sigEvtRef->evtRef);
        sigEvtRef->evtRef = NULL;
        free(sigEvtRef);
    }
    HASH_DELETE(hh, me->signalListRef, entryRef);
    free(entryRef);
}
//--------------------------------------------------------------------------------------------
void destroySignalList(EvtLoopRef me)
{
    assert(me != NULL);
    EvtListEntry_t* pStore = NULL;
    EvtListEntry_t* pEntry = NULL;
    HASH_ITER(hh, me->signalListRef, pEntry, pStore)
    {
        removeSignalElement(me, pEntry);
    }
    me->signalListRef = NULL;
}


//--------------------------------------------------------------------------------------------
void removeEPollElement(EvtLoopRef me, EvtListEntry_t* entryRef)
{
    assert(me != NULL);
    assert(entryRef != NULL);

    EvtLoopEventRef epollEvtRef = entryRef->evtRef;
    if(epollEvtRef != NULL)
    {
        epoll_ctl(me->epollFd, EPOLL_CTL_DEL, epollEvtRef->fd, NULL);
        free(epollEvtRef);
    }
    HASH_DELETE(hh, me->epollListRef, entryRef);
    free(entryRef);
}
//--------------------------------------------------------------------------------------------
void destroyEPollList(EvtLoopRef me)
{
    assert(me != NULL);
    EvtListEntry_t* pStore = NULL;
    EvtListEntry_t* pEntry = NULL;
    HASH_ITER(hh, me->epollListRef, pEntry, pStore)
    {
        removeEPollElement(me, pEntry);
    }
    me->epollListRef = NULL;
    if(me->epollFd != -1)
    {
        event_del(me->epollEvtRef);
        event_free(me->epollEvtRef);
        me->epollEvtRef = NULL;
        close(me->epollFd);
        me->epollFd = -1;
    }
}


//--------------------------------------------------------------------------------------------
void removeInotifyElement(EvtLoopRef me, EvtListEntry_t* entryRef)
{
    assert(me != NULL);
    assert(entryRef != NULL);

    EvtLoopEventRef inotifEvtRef = entryRef->evtRef;
    if(inotifEvtRef != NULL)
    {
        inotify_rm_watch(me->inotifyFd, inotifEvtRef->fd);
        free(inotifEvtRef);
    }
    HASH_DELETE(hh, me->inotifyListRef, entryRef);
    free(entryRef);
}
//--------------------------------------------------------------------------------------------
void destroyInotifyList(EvtLoopRef me)
{
    assert(me != NULL);
    EvtListEntry_t* pStore = NULL;
    EvtListEntry_t* pEntry = NULL;
    HASH_ITER(hh, me->inotifyListRef, pEntry, pStore)
    {
        removeInotifyElement(me, pEntry);
    }
    me->inotifyListRef = NULL;
    if(me->inotifyFd != -1)
    {
        event_del(me->inotifyEvtRef);
        event_free(me->inotifyEvtRef);
        me->inotifyEvtRef = NULL;
        close(me->inotifyFd);
        me->inotifyFd = -1;
    }
}


//--------------------------------------------------------------------------------------------
bool isTheLastDBusItemInList(EvtListEntry_t** pListRef, const char* dest, const char* path, const char* itf, const char* sig)
{
    return (NULL == iterateThroughListAndGetMatchedEntry(pListRef, dest, path, itf, sig, NULL));
}
//--------------------------------------------------------------------------------------------
void removeDBusElement(EvtLoopRef me, EvtListEntry_t* entryRef)
{
    assert(me != NULL);
    assert(entryRef != NULL);
    EvtLoopEventRef dbusEvtRef = entryRef->evtRef;
    assert(dbusEvtRef != NULL);

    HASH_DELETE(hh, me->dbusListRef, entryRef);
    const char* signalName = dbusEvtRef->dbusEvt.busId.signal;
    if(isTheLastDBusItemInList(&me->dbusListRef, dbusEvtRef->dbusEvt.busId.destination,
                               dbusEvtRef->dbusEvt.busId.objPath, dbusEvtRef->dbusEvt.busId.itf, signalName))
    {
        IUnixInitSys_unregisterDBusSignalMatch(me->initSysItfRef, dbusEvtRef->dbusEvt.busSlot);
        ILog_msg(me->logRef, LOG_LEVEL_DBG, "bus slot for signal=%s successfully unreferenced", signalName);
    }
    free(dbusEvtRef);
    free(entryRef);
}
//--------------------------------------------------------------------------------------------
void destroyDBusSignal(EvtLoopRef me)
{
    assert(me != NULL);
    EvtListEntry_t* pStore = NULL;
    EvtListEntry_t* pEntry = NULL;
    HASH_ITER(hh, me->dbusListRef, pEntry, pStore)
    {
        removeDBusElement(me, pEntry);
    }
    me->dbusListRef = NULL;
    if(me->dbusFd != -1)
    {
        event_del(me->dbusEvtRef);
        event_free(me->dbusEvtRef);
        me->dbusEvtRef = NULL;
        IUnixInitSys_deinitDBus(me->initSysItfRef);
        me->dbusFd = -1;
    }

}


//--------------------------------------------------------------------------------------------
uint32_t convertTimevalToMs(struct timeval* pTimeout)
{
    uint32_t timeoutMs = (uint32_t)((size_t)pTimeout->tv_sec * (size_t)MILLISECONDS_PER_SECOND);
    timeoutMs         += (uint32_t)((size_t)pTimeout->tv_usec / (size_t)MICROSECONDS_PER_MILLISECOND);
    return timeoutMs;
}
//--------------------------------------------------------------------------------------------
struct timeval convertMsToTimeval(uint32_t* pTimeoutMS)
{
    struct timeval timeout;
    timeout.tv_sec  = (__time_t)(*pTimeoutMS / MILLISECONDS_PER_SECOND);
    timeout.tv_usec = (__suseconds_t)((*pTimeoutMS % (uint32_t)MILLISECONDS_PER_SECOND) * MICROSECONDS_PER_MILLISECOND);
    return timeout;
}



/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
EvtLoopRef EvtLoop_create(ILogRef logRef, bool useExitSignals)
{
    EvtLoopRef me = (EvtLoop_t*)calloc(1, sizeof(EvtLoop_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating event loop instance");
    }
    me->epollFd        = -1;
    me->inotifyFd      = -1;
    me->dbusFd         = -1;
    me->logRef         = logRef;
    me->hasExitSignals = useExitSignals;
    me->evtBaseRef     = event_base_new();
    if(me->evtBaseRef == NULL)
    {
        Throw(EX_CREATE, "Creating a new event loop base");
    }
    int evtLoopFeatures  = event_base_get_features(me->evtBaseRef);
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "event loop supported features: 0x%X", evtLoopFeatures);

    me->signalListRef  = NULL;
    me->epollListRef   = NULL;
    me->inotifyListRef = NULL;
    me->dbusListRef    = NULL;

    if(me->hasExitSignals)
    {
        EvtLoop_registerSignalEvent(me, SIGINT,  &terminateApplication, me);
        EvtLoop_registerSignalEvent(me, SIGTERM, &terminateApplication, me);
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "Successfully created an event loop%s",
                                        (me->hasExitSignals) ? " (including termination signals)" : "");
    return me;
}
//--------------------------------------------------------------------------------------------
void EvtLoop_destroy(EvtLoopRef me)
{
    if(me != NULL)
    {
        destroyInotifyList(me);
        destroyEPollList(me);
        destroySignalList(me);
        destroyDBusSignal(me);

        if(me->evtBaseRef != NULL)
        {
            event_base_free(me->evtBaseRef);
            me->evtBaseRef = NULL;
        }
        free(me);
        me = NULL;
    }
}
//--------------------------------------------------------------------------------------------
void EvtLoop_reinit(EvtLoopRef me)
{
    assert(me != NULL);
    if(event_reinit(me->evtBaseRef) != 0)
    {
        Throw(EX_INIT, "Could not reinitialize event loop");
    }
}
//--------------------------------------------------------------------------------------------
int EvtLoop_run(EvtLoopRef me)
{
    assert(me != NULL);
    int ret;

    while(true)
    {
        ret = event_base_loop(me->evtBaseRef, 0);
        if(ret == -1)
        {
            ILog_msg(me->logRef, LOG_LEVEL_CRIT, "Internal error of libevent");
            break;
        }
        else if(ret == 0)
        {
            ILog_msg(me->logRef, LOG_LEVEL_DBG, "Event loop successfully exited");
            break;
        }
        else if(ret == 1)
        {
            ILog_msg(me->logRef, LOG_LEVEL_DBG, "No more events pending for event loop");
            break;
        }
        else
        {
            ILog_msg(me->logRef, LOG_LEVEL_CRIT, "Unknown error of libevent");
            break;
        }
    }

    return (ret >= 0) ? 0 : -1;
}
//--------------------------------------------------------------------------------------------
void EvtLoop_quit(EvtLoopRef me)
{
    assert(me != NULL);
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "User called event loop quit");
    event_base_loopexit(me->evtBaseRef, 0);
}




//--------------------------------------------------------------------------------------------
void EvtLoop_registerSignalEvent(EvtLoopRef evtLoopRef, int signalNbr, EvtLoopEventCallback cbFn, void *pCtx)
{
    assert(evtLoopRef != NULL);
    EvtLoopEventRef me = (EvtLoopEvent_t*)calloc(1, sizeof(EvtLoopEvent_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating event loop signal event");
    }
    me->fd            = signalNbr;
    me->evtType       = EVTLOOP_SIGNAL;
    me->contextRef    = pCtx;
    me->evtCallbackFn = cbFn;
    me->logRef        = evtLoopRef->logRef;
    me->evtRef        = event_new(evtLoopRef->evtBaseRef, me->fd, (EV_SIGNAL | EV_PERSIST), &signalEventCallbackFn, evtLoopRef);
    if(me->evtRef == NULL)
    {
        Throw(EX_CREATE, "Creating signal with libevent for %d", signalNbr);
    }

    createListEntryAndAddEventToList(&evtLoopRef->signalListRef, me);
    if(event_add(me->evtRef, NULL) != 0)
    {
        Throw(EX_ADD, "Adding created signal (%d) to event loop", me->fd);
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "event loop signal %d successfully created and added", me->fd);
}
//--------------------------------------------------------------------------------------------
void EvtLoop_unregisterSignalEvent(EvtLoopRef evtLoopRef, int signalNbr)
{
    assert(evtLoopRef != NULL);
    EvtListEntry_t* entryRef = NULL;
    HASH_FIND(hh, evtLoopRef->signalListRef, &signalNbr, (int)sizeof(signalNbr), entryRef);
    if(entryRef != NULL)
    {
        assert(entryRef->evtRef->fd == signalNbr);
        removeSignalElement(evtLoopRef, entryRef);
        ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "event loop signal %d successfully unregistered", signalNbr);
        return;
    }
    ILog_msg(evtLoopRef->logRef, LOG_LEVEL_NOTE, "event loop signal %d was not found for unregistering", signalNbr);
}


//--------------------------------------------------------------------------------------------
void EvtLoop_registerEPollEvent(EvtLoopRef evtLoopRef, int fd, EvtLoopEventCallback cbFn, bool isRegularFile, void *pCtx)
{
    assert(evtLoopRef != NULL);

    if(evtLoopRef->epollFd == -1)
    {
        evtLoopRef->epollFd = epoll_create1(0);
        if(evtLoopRef->epollFd < 0)
        {
            Throw(EX_CREATE, "Creating epoll: %s", strerror(errno));
        }

        evtLoopRef->epollEvtRef = event_new(evtLoopRef->evtBaseRef, evtLoopRef->epollFd,
                                            (EV_READ|EV_PERSIST), &epollEventCallbackFn, evtLoopRef);
        if(evtLoopRef->epollEvtRef == NULL)
        {
            Throw(EX_CREATE, "Creating epoll event with libevent");
        }
        if(event_add(evtLoopRef->epollEvtRef, NULL) != 0)
        {
            Throw(EX_ADD, "Adding created epoll to event loop");
        }
        ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "added epoll fd (%d) to event loop", evtLoopRef->epollFd);
    }

    EvtLoopEventRef me = (EvtLoopEvent_t*)calloc(1, sizeof(EvtLoopEvent_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating event loop epoll event");
    }
    me->fd            = fd;
    me->evtType       = EVTLOOP_EPOLL;
    me->contextRef    = pCtx;
    me->evtCallbackFn = cbFn;
    me->logRef        = evtLoopRef->logRef;

    struct epoll_event epollEvt;
    epollEvt.events  = (isRegularFile) ? EPOLLIN : EPOLLPRI;  // EPOLLIN = a file like /var/run/... | EPOLLPRI = e.g. regular GPIO
    epollEvt.data.fd = me->fd;
    if(epoll_ctl(evtLoopRef->epollFd, EPOLL_CTL_ADD, me->fd, &epollEvt) != 0)
    {
        Throw(EX_ADD, "Cannot add file descriptor (%d) to epoll: %s", me->fd, strerror(errno));
    }
    createListEntryAndAddEventToList(&evtLoopRef->epollListRef, me);
    evtLoopRef->epollEvtCnt++;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "event loop epoll event for fd=%d successfully created and added", me->fd);
}
//--------------------------------------------------------------------------------------------
void EvtLoop_unregisterEPollEvent(EvtLoopRef evtLoopRef, int fd)
{
    assert(evtLoopRef != NULL);
    EvtListEntry_t* entryRef = NULL;
    HASH_FIND(hh, evtLoopRef->epollListRef, &fd, (int)sizeof(fd), entryRef);
    if(entryRef != NULL)
    {
        assert(entryRef->evtRef->fd == fd);
        removeEPollElement(evtLoopRef, entryRef);
        evtLoopRef->epollEvtCnt--;
        ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "event loop epoll event with fd=%d successfully unregistered", fd);
        return;
    }
    ILog_msg(evtLoopRef->logRef, LOG_LEVEL_NOTE, "event loop epoll event with fd=%d was not found for unregistering", fd);
}


//--------------------------------------------------------------------------------------------
void EvtLoop_registerFilesystemChangeEvent(EvtLoopRef evtLoopRef, const char* filePath, EvtLoopEventCallback cbFn, void *pCtx)
{
    assert(evtLoopRef != NULL);
    if(evtLoopRef->inotifyFd == -1)
    {
        evtLoopRef->inotifyFd = inotify_init();
        if(evtLoopRef->inotifyFd < 0)
        {
            Throw(EX_CREATE, "Creating inotify: %s", strerror(errno));
        }
        evtLoopRef->inotifyEvtRef = event_new(evtLoopRef->evtBaseRef, evtLoopRef->inotifyFd,
                                              (EV_READ|EV_PERSIST), &filesystemChangedEventCallbackFn, evtLoopRef);
        if(evtLoopRef->inotifyEvtRef == NULL)
        {
            Throw(EX_CREATE, "Creating inotify event with libevent");
        }
        if(event_add(evtLoopRef->inotifyEvtRef, NULL) != 0)
        {
            Throw(EX_ADD, "Adding created inotify to event loop");
        }
        ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "added inotify fd (%d) to event loop", evtLoopRef->inotifyFd);
    }

    EvtLoopEventRef me = (EvtLoopEvent_t*)calloc(1, sizeof(EvtLoopEvent_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating event loop file change event");
    }

    uint32_t events   = IN_MODIFY;
    me->filePath      = filePath;
    me->evtType       = EVTLOOP_REGULAR_FILE;
    me->contextRef    = pCtx;
    me->evtCallbackFn = cbFn;
    me->logRef        = evtLoopRef->logRef;
    me->fd            = inotify_add_watch(evtLoopRef->inotifyFd, me->filePath, events);
    if(me->fd < 0)
    {
        Throw(EX_ADD, "Cannot add file %s to inotify: %s", me->filePath, strerror(errno));
    }
    createListEntryAndAddEventToList(&evtLoopRef->inotifyListRef, me);
    evtLoopRef->inotifyEvtCnt++;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "event loop file %s successfully created and added", me->filePath);
}
//--------------------------------------------------------------------------------------------
void EvtLoop_unregisterFileSystemChangeEvent(EvtLoopRef evtLoopRef, const char* filePath)
{
    assert(evtLoopRef != NULL);
    assert(filePath != NULL);

    EvtListEntry_t* pStore = NULL;
    EvtListEntry_t* pEntry = NULL;
    HASH_ITER(hh, evtLoopRef->inotifyListRef, pEntry, pStore)
    {
        if(strcmp(filePath, pEntry->evtRef->filePath) == 0)
        {
            removeInotifyElement(evtLoopRef, pEntry);
            evtLoopRef->inotifyEvtCnt--;
            ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "event loop file %s successfully unregistered", filePath);
            return;
        }
    }
    ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "event loop file %s was not found", filePath);
}


//--------------------------------------------------------------------------------------------
EvtLoopTimerRef EvtLoop_createAndRegisterTimer(EvtLoopRef evtLoopRef, EvtLoopEventCallback timeoutCbFn,
                                               uint32_t timeoutMS, bool isPeriodic, void *pCtx)
{
    assert(evtLoopRef != NULL);
    EvtLoopTimerRef me = (EvtLoopTimer_t*)calloc(1, sizeof(EvtLoopTimer_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating event loop timer instance");
    }
    me->evtRef = (EvtLoopEvent_t*)calloc(1, sizeof(EvtLoopEvent_t));
    if(me->evtRef == NULL)
    {
        Throw(EX_CREATE, "Creating event loop timer event");
    }
    me->evtRef->fd            = -1;
    me->evtRef->evtType       = EVTLOOP_TIMER;
    me->evtRef->contextRef    = pCtx;
    me->evtRef->evtCallbackFn = timeoutCbFn;
    me->evtRef->logRef        = evtLoopRef->logRef;
    me->isPeriodic            = isPeriodic;
    me->timeout               = convertMsToTimeval(&timeoutMS);
    me->evtRef->evtRef        = event_new(evtLoopRef->evtBaseRef, me->evtRef->fd, 0, &timerTimoutCallbackFn, me);
    if(me->evtRef->evtRef == NULL)
    {
        Throw(EX_CREATE, "Creating libevent timer");
    }
    ILog_msg(me->evtRef->logRef, LOG_LEVEL_DBG, "event loop timer with %ums timeout successfully created", timeoutMS);
    return me;
}
//--------------------------------------------------------------------------------------------
void EvtLoop_destroyTimer(EvtLoopTimerRef me)
{
    if(me != NULL)
    {
        if(me->isRunning)
        {
            EvtLoop_stopTimer(me);
        }
        event_free(me->evtRef->evtRef);
        free(me->evtRef);
        free(me);
    }
}
//--------------------------------------------------------------------------------------------
void EvtLoop_startTimer(EvtLoopTimerRef me)
{
    assert(me != NULL);
    if(event_add(me->evtRef->evtRef, &(me->timeout)) != 0)
    {
        Throw(EX_ADD, "Starting timer with timeout %ums", convertTimevalToMs(&me->timeout));
    }
    me->isRunning = true;
}
//--------------------------------------------------------------------------------------------
void EvtLoop_stopTimer(EvtLoopTimerRef me)
{
    assert(me != NULL);
    me->isRunning = false;
    if(event_del(me->evtRef->evtRef) != 0)
    {
        Throw(EX_REMOVE, "Stopping timer with timeout %ums", convertTimevalToMs(&me->timeout));
    }
}
//--------------------------------------------------------------------------------------------
bool EvtLoop_isTimerRunning(EvtLoopTimerRef me)
{
    assert(me != NULL);
    return me->isRunning;
}
//--------------------------------------------------------------------------------------------
void EvtLoop_setTimerTimeout(EvtLoopTimerRef me, uint32_t timeoutMS)
{
    assert(me != NULL);
    me->timeout = convertMsToTimeval(&timeoutMS);
}
//--------------------------------------------------------------------------------------------
uint32_t EvtLoop_getTimerTimeout(EvtLoopTimerRef me)
{
    assert(me != NULL);
    return convertTimevalToMs(&me->timeout);
}
//--------------------------------------------------------------------------------------------
void EvtLoop_setTimerPeriodicity(EvtLoopTimerRef me, bool isPeriodic)
{
    assert(me != NULL);
    me->isPeriodic = isPeriodic;
}
//--------------------------------------------------------------------------------------------
bool EvtLoop_isTimerPeriodic(EvtLoopTimerRef me)
{
    assert(me != NULL);
    return me->isPeriodic;
}


//--------------------------------------------------------------------------------------------
void EvtLoop_registerDBusSignal(EvtLoopRef evtLoopRef, IUnixInitSysRef initSysItfRef, const char* dbusService,
                                const char* dbusObjPath, const char* dbusItfName, const char* dbusSignalName,
                                const char* dbusSignalItem, EvtLoopEventCallback cbFn, void *pCtx)
{
    assert(evtLoopRef != NULL);
    assert(initSysItfRef != NULL);
    assert(dbusService != NULL);
    assert(dbusObjPath != NULL);
    assert(dbusItfName != NULL);
    assert(dbusSignalName != NULL);
    assert(dbusSignalItem != NULL);

    if(evtLoopRef->dbusFd == -1)
    {
        evtLoopRef->initSysItfRef = initSysItfRef;
        IUnixInitSys_initDBus(evtLoopRef->initSysItfRef);
        evtLoopRef->dbusFd = IUnixInitSys_getEventFileDescriptor(evtLoopRef->initSysItfRef);
        evtLoopRef->dbusEvtRef = event_new(evtLoopRef->evtBaseRef, evtLoopRef->dbusFd,
                                           (EV_READ|EV_PERSIST), &dbusEventCallbackFn, evtLoopRef);
        if(evtLoopRef->dbusEvtRef == NULL)
        {
            Throw(EX_CREATE, "Creating dbus event with libevent");
        }
        if(event_add(evtLoopRef->dbusEvtRef, NULL) != 0)
        {
            Throw(EX_ADD, "Adding created dbus event to event loop");
        }
        IUnixInitSys_registerDBusSignalCallback(evtLoopRef->initSysItfRef, &dbusSignalMatchCallbackFn, (void*)evtLoopRef);
        ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "added dbus event fd (%d) to event loop", evtLoopRef->dbusFd);
    }

    EvtLoopEventRef me = (EvtLoopEvent_t*)calloc(1, sizeof(EvtLoopEvent_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating event loop epoll event");
    }

    int uniqueDBusFd  = (int)(DBUS_FD_BASE_NBR + (int)(evtLoopRef->dbusEvtCnt) + 1);     // to have a unique identifier,
                                                                                         // the sum of a base number and
    me->fd            = uniqueDBusFd;                                                    // the event counter is used
    me->evtType       = EVTLOOP_DBUS;
    me->contextRef    = pCtx;
    me->evtCallbackFn = cbFn;
    me->logRef        = evtLoopRef->logRef;
    me->dbusEvt.initSysRef    = initSysItfRef;
    me->dbusEvt.evtItem       = dbusSignalItem;
    me->dbusEvt.busId.signal  = dbusSignalName;
    me->dbusEvt.busId.itf     = dbusItfName;
    me->dbusEvt.busId.objPath = dbusObjPath;
    me->dbusEvt.busId.destination = dbusService;

    createDBusListEntryAndAddEventToList(&evtLoopRef->dbusListRef, me);
    evtLoopRef->dbusEvtCnt++;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "event loop dbus signal/item=%s/%s successfully created and added",
                                        dbusSignalName, dbusSignalItem);
}
//--------------------------------------------------------------------------------------------
void EvtLoop_unregisterDBusSignal(EvtLoopRef evtLoopRef, IUnixInitSysRef initSysItfRef, const char* dbusService,
                                  const char* dbusObjPath, const char* dbusItfName, const char* dbusSignalName,
                                  const char* dbusSignalItem)
{
    assert(evtLoopRef != NULL);
    assert(dbusService != NULL);
    assert(dbusObjPath != NULL);
    assert(dbusItfName != NULL);
    assert(dbusSignalName != NULL);
    assert(dbusSignalItem != NULL);

    EvtListEntry_t* pEntry = iterateThroughListAndGetMatchedEntry(&evtLoopRef->dbusListRef, dbusService, dbusObjPath,
                                                                  dbusItfName, dbusSignalName, dbusSignalItem);
    if(pEntry != NULL)
    {
        removeDBusElement(evtLoopRef, pEntry);
        evtLoopRef->dbusEvtCnt--;
        return;
    }

    ILog_msg(evtLoopRef->logRef, LOG_LEVEL_DBG, "event loop dbus signal/item %s/%s was not found", dbusSignalName, dbusSignalItem);
}



/************************************END OF FILE*********************************************/
