/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * Exception Handler for POSIX systems
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : exception-posix.c
 * CREATION DATE : Mar 10, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <unistd.h>
#include <signal.h>
#include <pthread.h>
#include <nmapp/exception/exception.h>
#include <nmapp/logging/ILog.h>

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

typedef struct __ThreadStruct
{
    pthread_t id;
} ThreadStruct_t;

static ExceptionInstance_t exceptionInstances[EXCEPTION_MAX_INSTANCES];
static ThreadStruct_t      threads[EXCEPTION_MAX_INSTANCES];
static sig_atomic_t        isInitialized = false;
static pthread_mutex_t     getInstanceMutex = PTHREAD_MUTEX_INITIALIZER;


/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
ExceptionInstance_t* Exception_getInstance(void)
{
    ExceptionInstance_t* exInstance = NULL;
    int32_t unusedInstance = -1;
    pthread_mutex_lock(&getInstanceMutex);
    if(isInitialized == false)
    {
        memset(exceptionInstances, 0, sizeof(ExceptionInstance_t) * EXCEPTION_MAX_INSTANCES);
        isInitialized = true;
    }
    int32_t i;
    pthread_t pTh = pthread_self();
    for(i = 0; i < EXCEPTION_MAX_INSTANCES; i++)
    {
        exceptionInstances[i].thread = &threads[i];
        unusedInstance = exceptionInstances[i].isUsed == false ? i : unusedInstance;
        if(pthread_equal(exceptionInstances[i].thread->id, pTh) != 0)
        {
            exInstance = &exceptionInstances[i];
            break;
        }
    }
    //no exception instance found (thread not yet registered)
    if(exInstance == NULL)
    {
        //all exception instances used -> cleanup or error handling
        if(unusedInstance < 0)
        {
            //TODO: add cleanup procedure for no longer existing threads?
            //problem see: https://stackoverflow.com/a/9870708
            ILog_msg(NULL, LOG_LEVEL_CRIT, "No more exception instances available");
            pthread_mutex_unlock(&getInstanceMutex);
            assert(false);
            return NULL;
        }
        //space available -> register
        exceptionInstances[unusedInstance].thread->id = pTh;
        exceptionInstances[unusedInstance].isUsed = true;
        exInstance = &exceptionInstances[unusedInstance];
    }
    pthread_mutex_unlock(&getInstanceMutex);

    return exInstance;
}
//--------------------------------------------------------------------------------------------
void handleExceptionError(ExceptionInstance_t* exInst)
{
    // set ILogRef to NULL to use the system logger
    ILog_msg(NULL, LOG_LEVEL_EMCY, "Exeption Error: uncaught Throw (missing global Try/Catch?)");
    ILog_msg(NULL, LOG_LEVEL_EMCY, "Exeption Error: current.line = %d", exInst->exception.current.line);
    ILog_msg(NULL, LOG_LEVEL_EMCY, "Exeption Error: current.file = %s", exInst->exception.current.file);
    ILog_msg(NULL, LOG_LEVEL_EMCY, "Exeption Error: current.function = %s", exInst->exception.current.function);
    ILog_msg(NULL, LOG_LEVEL_EMCY, "Exeption Error: current.id = %d", exInst->exception.current.id);
    ILog_msg(NULL, LOG_LEVEL_EMCY, "Exeption Error: current.msg = %s", exInst->exception.current.msg);
    ILog_msg(NULL, LOG_LEVEL_EMCY, "Exeption Error: going to assert for SegFault prevention...");
    sleep(1);
    assert(false);
}

#ifdef __cplusplus
}
#endif

/************************************END OF FILE*********************************************/
