/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * FileOperation MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : FileOperation.c
 * CREATION DATE : Apr 2, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdlib.h>                      // for calloc, free
#include <stdio.h>                       // for rename, remove
#include <unistd.h>                      // for access, ftruncate
#include <errno.h>                       // for errno
#include <sys/stat.h>                    // for struct stat

#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>

#include "FileOperation-private.h"

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/
#define FILEOP_BUFSIZE 8192


//--------------------------------------------------------------------------------------------
void readFullLine(FileOperationRef me, char** pBuffer, uint32_t size, uint32_t lineNbr, uint32_t tgtLine)
{
    assert(me != NULL);
    assert(pBuffer != NULL);

    unsigned int humanLineNbr = lineNbr + 1;
    size_t lineLen = 0;
    char*  line = *pBuffer;
    do
    {
        memset(line, 0, size);
        if(fgets(line, (int)size, me->file) == NULL)
        {
            if(feof(me->file) != 0)
            {
                Throw(EX_READ, "End of file reached before reading line nbr=%u", (unsigned int)tgtLine);
            }
            Throw(EX_READ, "Failed reading line %u (demanded=%u) from %s: %s", humanLineNbr, (unsigned int)tgtLine,
                                                                               me->fileName, strerror(errno));
        }
        lineLen =  strlen(line);
    } while(line[lineLen-1] != '\n');
}

/*********************************************************************************************
 *        INTERFACE REALIZATIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
bool isExist(IFileOperationImplRef implRef, const char* fileName)
{
    assert(fileName != NULL);
    return (access(fileName, F_OK) == 0);
}
//--------------------------------------------------------------------------------------------
bool isDirectory(IFileOperationImplRef implRef, const char* path)
{
    assert(path != NULL);
    struct stat pathStat;
    if(stat(path, &pathStat) != 0)
    {
        Throw(EX_READ, "isDirectory - cannot stat '%s': %s", path, strerror(errno));
    }
    return S_ISDIR(pathStat.st_mode);
}
//--------------------------------------------------------------------------------------------
int getFiledescriptor(IFileOperationImplRef implRef)
{
    assert(implRef != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    int fd = fileno(me->file);
    if(fd < 0)
    {
        Throw(EX_CONVERSION, "Filed getting file descriptor for file %s: %s", me->fileName, strerror(errno));
    }
    return fd;
}
//--------------------------------------------------------------------------------------------
bool isFileOpen(IFileOperationImplRef implRef)
{
    assert(implRef != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    return (me->file != NULL);
}
//--------------------------------------------------------------------------------------------
void openFile(IFileOperationImplRef implRef, const char* fileName, const char* modes)
{
    assert(implRef != NULL);
    assert(fileName != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    if(me->file != NULL)
    {
        Throw(EX_OPEN, "A file (%s) is already openened", me->fileName);
    }
    me->file = fopen(fileName, modes);
    if(NULL == me->file)
    {
        Throw(EX_OPEN, "Couldn't open file %s: %s", fileName, strerror(errno));
    }
    me->fileName = fileName;
}
//--------------------------------------------------------------------------------------------
void closeFile(IFileOperationImplRef implRef)
{
    assert(implRef != NULL);
    if(!isFileOpen(implRef))
    {
        // no file to close
        return;
    }

    FileOperationRef me = (FileOperationRef)implRef;
    if(fclose(me->file) != 0)
    {
        Throw(EX_CLOSE, "Couldn't close file %s: %s", me->fileName, strerror(errno));
    }
    me->file = NULL;
    me->fileName = NULL;
}
//--------------------------------------------------------------------------------------------
const char* getFileName(IFileOperationImplRef implRef)
{
    assert(implRef != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    return me->fileName;
}
//--------------------------------------------------------------------------------------------
void writeToFile(IFileOperationImplRef implRef, const char* data, uint32_t size)
{
    assert(implRef != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    if(me->file == NULL)
    {
        Throw(EX_WRITE, "There is no opened file to write to");
    }
    long offset = 0;
    if(fseek(me->file, offset, SEEK_END) != 0)
    {
        Throw(EX_READ, "Could not go to the end to write to %s: %s", me->fileName, strerror(errno));
    }
    if(fwrite(data, size, 1, me->file) < 1)
    {
        Throw(EX_WRITE, "Could not write data to %s: %s", me->fileName, strerror(errno));
    }
}
//--------------------------------------------------------------------------------------------
void writeUpdateToFile(IFileOperationImplRef implRef, bool withTruncate, uint32_t atOffset, const char* data, uint32_t size)
{
    assert(implRef != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    if(me->file == NULL)
    {
        Throw(EX_WRITE, "There is no opened file to write an update to");
    }
    if(fseek(me->file, (long int)atOffset, SEEK_SET) != 0)
    {
        Throw(EX_READ, "Could not go to the offset %lu to write an update to %s: %s", (unsigned long int)atOffset,
                                                                                     me->fileName, strerror(errno));
    }
    if(fwrite(data, size, 1, me->file) < 1)
    {
        Throw(EX_WRITE, "Could not write data to %s: %s", me->fileName, strerror(errno));
    }
    off_t newLength = (off_t)(atOffset + size);
    int fd = getFiledescriptor(implRef);
    if(ftruncate(fd, newLength) != 0)
    {
        Throw(EX_WRITE, "Could not trim file %s to updated data: %s", me->fileName, strerror(errno));
    }
}
//--------------------------------------------------------------------------------------------
void readFromFile(IFileOperationImplRef implRef, char* buffer, uint32_t size, uint32_t offset)
{
    assert(implRef != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    if(me->file == NULL)
    {
        Throw(EX_WRITE, "There is no opened file to read from");
    }
    if(fseek(me->file, (long)offset, SEEK_SET) != 0)
    {
        Throw(EX_READ, "Could not go to the offset of %uBytes to read from %s: %s", (unsigned int)offset, me->fileName, strerror(errno));
    }
    size_t readSize = fread(buffer, 1, size, me->file);
    if((ferror(me->file) != 0) || ((readSize == 0) && (feof(me->file) != 0)))
    {
        Throw(EX_READ, "Could not read from %s: %s", me->fileName, strerror(errno));
    }
}
//--------------------------------------------------------------------------------------------
void readLineFromFile(IFileOperationImplRef implRef, uint32_t lineNbr, char* buffer, uint32_t size)
{
    assert(implRef != NULL);
    FileOperationRef me = (FileOperationRef)implRef;
    if(me->file == NULL)
    {
        Throw(EX_WRITE, "There is no opened file to read a line from");
    }
    // Set cursor to beginning of file:
    long offset = 0;
    if(fseek(me->file, offset, SEEK_SET) != 0)
    {
        Throw(EX_READ, "Could not go to the end to read from %s: %s", me->fileName, strerror(errno));
    }
    uint32_t actLine;
    for(actLine = 0; actLine < lineNbr; actLine++)
    {
        readFullLine(me, &buffer, size, actLine, lineNbr);
    }
    // buffer holds the string of the desired line number
}
//--------------------------------------------------------------------------------------------
void moveFile(IFileOperationImplRef implRef, const char* srcFileName, const char* dstFileName)
{
    assert(srcFileName != NULL);
    assert(dstFileName != NULL);
    if(rename(srcFileName, dstFileName) != 0)
    {
        Throw(EX_WRITE, "Moving/Renaming file '%s' to '%s': %s", srcFileName, dstFileName, strerror(errno));
    }
}
//--------------------------------------------------------------------------------------------
void copyFile(IFileOperationImplRef implRef, const char* srcFileName, const char* dstFileName)
{
    assert(srcFileName != NULL);
    assert(dstFileName != NULL);

    char    buffer[FILEOP_BUFSIZE];
    size_t  n;
    FILE   *fsrc;
    FILE   *fdest;
    if((fsrc = fopen(srcFileName, "rb")) == NULL)
    {
        Throw(EX_OPEN, "cannot open source file %s for copying: %s", srcFileName, strerror(errno));
    }
    if((fdest = fopen(dstFileName, "wb")) == NULL)
    {
        Throw(EX_OPEN, "cannot open/create destination file %s for copying: %s", dstFileName, strerror(errno));
    }

    while((n = fread(buffer, sizeof(char), sizeof(buffer), fsrc)) > 0)
    {
        if(fwrite(buffer, sizeof(char), n, fdest) != n)
        {
            Throw(EX_WRITE, "Copying file '%s' to '%s': %s", srcFileName, dstFileName, strerror(errno));
        }
    }
    fclose(fsrc);
    fclose(fdest);
}
//--------------------------------------------------------------------------------------------
void deleteFilePath(IFileOperationImplRef implRef, const char* filePath)
{
    assert(filePath != NULL);
    if(remove(filePath) != 0)
    {
        Throw(EX_REMOVE, "Deleting file path '%s': %s", filePath, strerror(errno));
    }
}
//--------------------------------------------------------------------------------------------
void createDirectory(IFileOperationImplRef implRef, const char* dirPath, uint32_t permissions)
{
    assert(dirPath != NULL);
    if(mkdir(dirPath, (mode_t)permissions) != 0)
    {
        Throw(EX_CREATE, "Create directory '%s': %s", dirPath, strerror(errno));
    }
}


/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
FileOperationRef FileOperation_create(ILogRef logRef)
{
    FileOperationRef me = (FileOperationPrivate_t*)calloc(1, sizeof(FileOperationPrivate_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating FileOperation failed");
    }
    me->file     = NULL;
    me->fileName = NULL;
    me->logRef   = logRef;
    me->itf      = IFileOperation_ctor(me, &isFileOpen, &openFile, &closeFile, &getFileName,
                                           &writeToFile, &writeUpdateToFile, &readFromFile, &readLineFromFile,
                                           &moveFile, &copyFile, &deleteFilePath, &createDirectory,
                                           &isExist, &isDirectory,
                                           &getFiledescriptor);
    return me;
}
//--------------------------------------------------------------------------------------------
void FileOperation_destroy(FileOperationRef me)
{
    if(me != NULL)
    {
        closeFile((IFileOperationImplRef)me);
        IFileOperation_destroy(me->itf);
        free(me);
    }
}
//--------------------------------------------------------------------------------------------
IFileOperationRef FileOperation_getItfRef(FileOperationRef me)
{
    assert(me != NULL);
    return me->itf;
}
