/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * IFileOperation (realization part of the interface)
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : IFileOperation.c
 * CREATION DATE : Apr 2, 2020
 * TARGET        : generic
 *
 *********************************************************************************************/

#include <stdlib.h>                         // for calloc, free, ...
#include <nmapp/types/nm-types.h>
#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include "IFileOperation-protected.h"


/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/


/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
IFileOperationRef IFileOperation_ctor(IFileOperationImplRef implRef,
                                      bool (*isFileOpen)(IFileOperationImplRef implRef),
                                      void (*openFile)(IFileOperationImplRef implRef, const char* fileName, const char* modes),
                                      void (*closeFile)(IFileOperationImplRef implRef),
                                      const char* (*getFileName)(IFileOperationImplRef implRef),
                                      void (*writeToFile)(IFileOperationImplRef implRef, const char* data, uint32_t size),
                                      void (*writeUpdateToFile)(IFileOperationImplRef implRef, bool withTruncate, uint32_t atOffset, const char* data, uint32_t size),
                                      void (*readFromFile)(IFileOperationImplRef implRef, char* buffer, uint32_t size, uint32_t offset),
                                      void (*readLineFromFile)(IFileOperationImplRef implRef, uint32_t lineNbr, char* buffer, uint32_t size),
                                      void (*moveFile)(IFileOperationImplRef implRef, const char* srcFileName, const char* dstFileName),
                                      void (*copyFile)(IFileOperationImplRef implRef, const char* srcFileName, const char* dstFileName),
                                      void (*deleteFilePath)(IFileOperationImplRef implRef, const char* filePath),
                                      void (*createDirectory)(IFileOperationImplRef implRef, const char* dirPath, uint32_t permissions),
                                      bool (*isExist)(IFileOperationImplRef implRef, const char* fileName),
                                      bool (*isDirectory)(IFileOperationImplRef implRef, const char* path),
                                      int  (*getFiledescriptor)(IFileOperationImplRef implRef))
{
    IFileOperationRef me = (IFileOperationRef)calloc(1, sizeof(IFileOperationProtected_t));
    assert(me != NULL);

    me->myImpl = implRef;
    me->isFileOpen  = isFileOpen;
    me->openFile    = openFile;
    me->closeFile   = closeFile;
    me->getFileName = getFileName;
    me->getFiledescriptor = getFiledescriptor;

    me->writeToFile       = writeToFile;
    me->writeUpdateToFile = writeUpdateToFile;
    me->readFromFile      = readFromFile;
    me->readLineFromFile  = readLineFromFile;

    me->moveFile        = moveFile;
    me->copyFile        = copyFile;
    me->deleteFilePath  = deleteFilePath;
    me->createDirectory = createDirectory;
    me->isExist         = isExist;
    me->isDirectory     = isDirectory;

    assert(me->isFileOpen != NULL);
    assert(me->openFile != NULL);
    assert(me->closeFile != NULL);
    assert(me->getFileName != NULL);
    assert(me->getFiledescriptor != NULL);
    assert(me->writeToFile != NULL);
    assert(me->writeUpdateToFile != NULL);
    assert(me->readFromFile != NULL);
    assert(me->readLineFromFile != NULL);
    assert(me->isExist != NULL);
    assert(me->isDirectory != NULL);
    assert(me->moveFile != NULL);
    assert(me->copyFile != NULL);
    assert(me->deleteFilePath != NULL);
    assert(me->createDirectory != NULL);
    return me;
}
//--------------------------------------------------------------------------------------------
void IFileOperation_destroy(IFileOperationRef me)
{
    if(me != NULL)
    {
        free(me);
    }
}

//--------------------------------------------------------------------------------------------
bool IFileOperation_isFileOpen(IFileOperationRef me)
{
    assert(me != NULL);
    return me->isFileOpen(me->myImpl);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_open(IFileOperationRef me, const char* fileName, const char* modes)
{
    assert(me != NULL);
    me->openFile(me->myImpl, fileName, modes);
}
//--------------------------------------------------------------------------------------------
const char* IFileOperation_getFileName(IFileOperationRef me)
{
    assert(me != NULL);
    return me->getFileName(me->myImpl);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_close(IFileOperationRef me)
{
    assert(me != NULL);
    me->closeFile(me->myImpl);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_writeToFile(IFileOperationRef me, const char* data, uint32_t size)
{
    assert(me != NULL);
    me->writeToFile(me->myImpl, data, size);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_writeUpdateToFile(IFileOperationRef me, bool withTruncate,
                                      uint32_t atOffset, const char* data, uint32_t size)
{
    assert(me != NULL);
    me->writeUpdateToFile(me->myImpl, withTruncate, atOffset, data, size);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_readFromFile(IFileOperationRef me, char* buffer, uint32_t size, uint32_t offset)
{
    assert(me != NULL);
    me->readFromFile(me->myImpl, buffer, size, offset);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_readLineFromFile(IFileOperationRef me, uint32_t lineNbr, char* line, uint32_t size)
{
    assert(me != NULL);
    me->readLineFromFile(me->myImpl, lineNbr, line, size);
}

//--------------------------------------------------------------------------------------------
void IFileOperation_move(IFileOperationRef me, const char* srcFileName, const char *dstFileName)
{
    assert(me != NULL);
    me->moveFile(me->myImpl, srcFileName, dstFileName);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_copy(IFileOperationRef me, const char* srcFileName, const char* dstFileName)
{
    assert(me != NULL);
    me->copyFile(me->myImpl, srcFileName, dstFileName);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_delete(IFileOperationRef me, const char* filePath)
{
    assert(me != NULL);
    me->deleteFilePath(me->myImpl, filePath);
}
//--------------------------------------------------------------------------------------------
void IFileOperation_createDirectory(IFileOperationRef me, const char* dirPath, uint32_t permissions)
{
    assert(me != NULL);
    me->createDirectory(me->myImpl, dirPath, permissions);
}
//--------------------------------------------------------------------------------------------
bool IFileOperation_isExist(IFileOperationRef me, const char* fileName)
{
    assert(me != NULL);
    return me->isExist(me->myImpl, fileName);
}
//--------------------------------------------------------------------------------------------
bool IFileOperation_isDirectory(IFileOperationRef me, const char* path)
{
    assert(me != NULL);
    return me->isDirectory(me->myImpl, path);
}
//--------------------------------------------------------------------------------------------
int IFileOperation_getFiledescriptor(IFileOperationRef me)
{
    assert(me != NULL);
    return me->getFiledescriptor(me->myImpl);
}

