/*********************************************************************************************
 * (c) COPYRIGHT 2021 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * IoctlOperation MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : IoctlOperation.c
 * CREATION DATE : Oct 21, 2021
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdlib.h>                      // for calloc, free
#include <fcntl.h>                       // for open
#include <unistd.h>                      // for close
#include <errno.h>                       // for errno
#include <string.h>                      // for strerror
#include <linux/major.h>                 // for mmc major number
#include <linux/mmc/ioctl.h>             // for mmc specific ioctl

#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>

#include "IoctlOperation-private.h"

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

/*********************************************************************************************
 *        INTERFACE REALIZATIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
static int ioctlRequest(IIoctlOperationImplRef implRef, const char* fileName, unsigned long request, void* argp)
{
    assert(implRef != NULL);
    assert(fileName != NULL);
    IoctlOperationRef me = (IoctlOperationRef)implRef;
    int fd = open(fileName, O_NONBLOCK | O_RDWR);
    if (fd < 0)
    {
        Throw(EX_OPEN, "Could not open %s: %s", fileName, strerror(errno));
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "calling ioctl on %s, request=%lu/0x%lX", fileName, request, request);
    int ioctlRet = ioctl(fd, request, argp);
    close(fd);
    return ioctlRet;
}
//--------------------------------------------------------------------------------------------
static void mmcIoctlRequest(IIoctlOperationImplRef implRef, const char* devPath, void* pIocCmd, void* pData)
{
    assert(implRef != NULL);
    assert(pIocCmd != NULL);
    assert(pData != NULL);

    IoctlOperationRef me = (IoctlOperationRef)implRef;
    struct mmc_ioc_cmd requestData = *((struct mmc_ioc_cmd*)(pIocCmd));

    int fd = open(devPath, O_NONBLOCK | O_RDWR);
    if (fd < 0)
    {
        Throw(EX_OPEN, "Could not open %s: %s", devPath, strerror(errno));
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "set data for mmc ioc command for device %s", devPath);
    mmc_ioc_cmd_set_data(requestData, (uint8_t*)pData);
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "calling mmc ioctl on device %s", devPath);
    int ioctlRet = ioctl(fd, MMC_IOC_CMD, &requestData);
    if(ioctlRet != 0)
    {
        close(fd);
        Throw(EX_TRANSMIT, "ioctl MMC_IOC_CMD: %s", strerror(errno));
    }
    close(fd);
}

/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
IoctlOperationRef IoctlOperation_create(ILogRef logRef)
{
    IoctlOperationRef me = (IoctlOperationPrivate_t*)calloc(1, sizeof(IoctlOperationPrivate_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Creating IoctlOperation failed");
    }
    me->logRef = logRef;
    me->itf = IIoctlOperation_ctor(me, &ioctlRequest, &mmcIoctlRequest);
    return me;
}
//--------------------------------------------------------------------------------------------
void IoctlOperation_destroy(IoctlOperationRef me)
{
    if(me != NULL)
    {
        IIoctlOperation_destroy(me->itf);
        free(me);
    }
}
//--------------------------------------------------------------------------------------------
IIoctlOperationRef IoctlOperation_getItfRef(IoctlOperationRef me)
{
    assert(me != NULL);
    return me->itf;
}
