/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * System Logger Command Line
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : SysLogger-cmdline.c
 * CREATION DATE : Mar 26, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <libgen.h>    // for basename
#include <stdlib.h>    // for atoi
#include <stdio.h>     // for printf

#include <nmapp/logging/SysLogger-cmdline.h>
#include <nmapp/exception/exception.h>

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

#define LOGGERHELP_OPT_LEVEL    "--loglevel"
#define LOGGERHELP_OPT_DISABLE  "--disable-syslog"
#define LOGGERHELP_OPT_STDOUT   "--enable-stdout"

#define LOGGERCFG_OPTDELIM      '='
#define LOGGERCFG_CFGDELIM      ','
#define LOGGERCFG_LEVELDELIM    '.'

SysLoggersCmdLine_t cmdLineLoggers;


/*********************************************************************************************
 *        LOCAL Helper FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
LogLevel_t extractLogLevel(char* strLevel)
{
    int theLevel;

    if(isdigit(*strLevel) == 0)
    {
        Throw(EX_MALFORMED, "Log level is not a digit char=%c", *strLevel);
    }
    theLevel = atoi(strLevel);
    if(theLevel >= (int)LOG_LEVEL_MAX)
    {
        Throw(EX_OUT_OF_RANGE, "The log level is out of a valid range (%d ... %d), level=%d",
                                (int)LOG_LEVEL_EMCY, (int)LOG_LEVEL_DBG, theLevel);
    }
    return (LogLevel_t)theLevel;
}
//--------------------------------------------------------------------------------------------
char* extractGeneralLogLevelAndReturnNextStringBeginning(SysLoggersCmdLine_t* me, char* strLevel)
{
    assert(me != NULL);
    char* token = strLevel;

    if(isdigit(*token))
    {
        me->maxLogLevel = extractLogLevel(token);
        // Sanity check for the next character:
        // must either have a string termination or a LOGGERCFG_CFGDELIM to separate general loglevel from loggers name
        token++;
        if((*token == 0) || (*token != LOGGERCFG_CFGDELIM))
        {
            Throw(EX_MALFORMED, "The log level configuration is incorrect formatted");
        }
        token++;
    }
    return token;
}
//--------------------------------------------------------------------------------------------
size_t getLoggerCfgSize(char* token)
{
    char *delim = strchr(token, LOGGERCFG_CFGDELIM);
    return (delim == NULL) ? strlen(token) : (size_t)(delim - token);
}
//--------------------------------------------------------------------------------------------
size_t getLoggerNameSize(char* token)
{
    char *delim = strchr(token, LOGGERCFG_LEVELDELIM);
    return (delim == NULL) ? 0 : (size_t)(delim - token);
}
//--------------------------------------------------------------------------------------------
void extractAdditionalLoggerConfigs(SysLoggersCmdLine_t* me, char* strLevel)
{
    assert(me != NULL);

    char* token = strLevel;
    int   newLoggerIdx = 0;

    // Per logger configuration
    while(*token != 0)
    {
        char*      loggerCfgStart = token;
        size_t     loggerCfgSize  = getLoggerCfgSize(token);
        size_t     loggerNameSize = getLoggerNameSize(token);

        // Perform some sanity checks
        if(newLoggerIdx >= SYSLOG_MAX_NBR_OF_LOGGERS)
        {
            Throw(EX_OUT_OF_RANGE, "The number of loggers are reached");
        }
        if(loggerCfgSize == 0)
        {
            // no valid config -> abort
            return;
        }

        // there is a logger name without loglevel given -> skip this logger
        if((loggerNameSize == 0) || (loggerCfgSize < loggerNameSize))
        {
            token += (loggerCfgSize + 1);
            continue;
        }

        // set token to level
        token += (loggerNameSize + 1);
        LogLevel_t loggerLevel = extractLogLevel(token++);

        // Save extracted logger configuration
        size_t loggerNameSizeTrimmed = MIN((SYSLOG_MAX_LOGGER_NAME_LENGTH - 1), loggerNameSize);
        me->loggers[newLoggerIdx].maxLogLevel = loggerLevel;
        memcpy(me->loggers[newLoggerIdx].loggerName, loggerCfgStart, loggerNameSizeTrimmed);
        me->loggers[newLoggerIdx].loggerName[loggerNameSizeTrimmed] = '\0';

        // Next logger
        token++;
        newLoggerIdx++;
    }
}
//--------------------------------------------------------------------------------------------
void parseLogLevelString(SysLoggersCmdLine_t* me, char* strLogLevel)
{
    assert(me != NULL);

    Exception_t ex;
    char* token = strLogLevel;

    Try
    {
        token = extractGeneralLogLevelAndReturnNextStringBeginning(me, token);
        extractAdditionalLoggerConfigs(me, token);
    }
    Catch(ex)
    {
        if(ex.id == EX_MALFORMED)
        {
            // There is a potential issue with the configuration string format
            // --> we do currently nothing
        }
        if(ex.id == EX_OUT_OF_RANGE)
        {
            // Not all loggers could be configured due to exceeding number of supported loggers or
            // due to exceeding log level
            // --> we do currently nothing
        }
    }
}

/*********************************************************************************************
 *        LOCAL Main FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
void setDefaults(SysLoggersCmdLine_t* me)
{
    assert(me != NULL);
    memset(me, 0, sizeof(SysLoggersCmdLine_t));
    me->useStdout   = false;
    me->useSysLog   = true;
    me->maxLogLevel = LOG_LEVEL_NOTE;
}
//--------------------------------------------------------------------------------------------
void extractAndSetApplicationName(SysLoggersCmdLine_t* me, int32_t argc, char * argv[])
{
    assert(me != NULL);
    if(argc > 0)
    {
        me->appName = basename(argv[0]);
    }
}
//--------------------------------------------------------------------------------------------
void parseCommandLine(SysLoggersCmdLine_t* me, int32_t argc, char * argv[])
{
    assert(me != NULL);
    int argIdx = 0;

    while(++argIdx < argc)
    {
        char* delim = strchr(argv[argIdx], LOGGERCFG_OPTDELIM);
        if(delim != NULL)
        {
            char* argValue = delim + 1;
            size_t argSize = (size_t)(delim - argv[argIdx]);
            if(*argValue == 0)
            {
                // no argument value given --> continue with next argument
                continue;
            }

            if(strncmp(argv[argIdx], LOGGERHELP_OPT_LEVEL, argSize) == 0)
            {
                parseLogLevelString(me, argValue);
            }
        }
        if(!strcmp(argv[argIdx], LOGGERHELP_OPT_DISABLE))
        {
            me->useSysLog = false;
        }
        if(!strcmp(argv[argIdx], LOGGERHELP_OPT_STDOUT))
        {
            me->useStdout = true;
        }
    }
}

/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
SysLoggersCmdLine_t SysLogCmdLine_parse(int argc, char* argv[])
{
    SysLoggersCmdLine_t* me = &cmdLineLoggers;

    setDefaults(me);
    extractAndSetApplicationName(me, argc, argv);
    parseCommandLine(me, argc, argv);
    return *me;
}
//--------------------------------------------------------------------------------------------
bool SysLogCmdLine_isValidCommandLineOption(const char* opt)
{
    char optionStr[256] = "";
    strncpy(optionStr, opt, 255);
    char* delim = strchr(optionStr, LOGGERCFG_OPTDELIM);
    if(delim != NULL)
    {
        *delim = '\0';
    }
    return ((strcmp(optionStr, LOGGERHELP_OPT_LEVEL) == 0)   ||
            (strcmp(optionStr, LOGGERHELP_OPT_DISABLE) == 0) ||
            (strcmp(optionStr, LOGGERHELP_OPT_STDOUT) == 0));
}
//--------------------------------------------------------------------------------------------
void SysLogCmdLine_printHelp(void)
{
    fprintf(stdout,
        "SysLogger    OPTIONS:\n"
        "             %s%cn          Set the max application log level (used for all logger\n"
        "                                   instances as default) to n (0=emcy, 7=dbg). If comma separated\n"
        "                                   list of separate logger instances is provided after this\n"
        "                                   number, the log level for each such instance will be overruled\n"
        "                                   accordingly (e.g. --loglevel=7,evtloop.5,fileOp.6).\n"
        "             %s      disable the log output to syslog\n"
        "             %s       enable output on stdout (e.g. for debugging purposes)\n\n"
        "SysLogger Examples:\n"
        "             prog-name %s\n"
        "             prog-name %s %s\n"
        "             prog-name %s%c6,config.5,serial.7\n"
        "\n\n", LOGGERHELP_OPT_LEVEL, LOGGERCFG_OPTDELIM, LOGGERHELP_OPT_DISABLE, LOGGERHELP_OPT_STDOUT,
                LOGGERHELP_OPT_DISABLE,
                LOGGERHELP_OPT_DISABLE, LOGGERHELP_OPT_STDOUT,
                LOGGERHELP_OPT_LEVEL, LOGGERCFG_OPTDELIM);
}

