/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * System Logger
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : SysLogger.c
 * CREATION DATE : Mar 24, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <getopt.h>

#include <nmapp/exception/exception.h>
#include <nmapp/logging/SysLogger.h>
#include <nmapp/logging/SysLogger-cmdline.h>
#include <nmapp/logging/ILog-protected.h>
#include <nmapp/sys/UnixInitSys.h>

#include "SysLogger-private.h"

#ifdef __cplusplus
extern "C" {
#endif

/*********************************************************************************************
 *        LOCAL DECLARATIONS
 *********************************************************************************************/

// Constants:
//-----------------------------------------
#define SYSLOG_MSG_STDOUT_FORMAT_STR    "%s:" SYSLOG_MSG_FORMAT_STR

//log string color for ANSII terminal (e.g. for routing log messages to stdout)
#define ANSI_COLOR_RED                  "\033[1;31m"
#define ANSI_COLOR_YELLOW               "\033[1;33m"
#define ANSI_COLOR_BLUE                 "\e[34m"
#define ANSI_COLOR_WHITE                "\033[1;37m"
#define ANSI_COLOR_GRAY                 "\033[39m"
#define ANSI_COLOR_END                  "\e[0m"

const int32_t SYSLOG_FACILITY[SYSLOG_FACILITY_MAX] =
{
    [SYSLOG_FACILITY_ROUTER] = LOG_LOCAL0,     //from syslog.h -> used for router log messages
};

const char* SYSLOG_FACILITY_NAME[SYSLOG_FACILITY_MAX] =
{
    [SYSLOG_FACILITY_ROUTER] = "ROUTER",
};



// Variables:
//-----------------------------------------
SysLoggerRef sysLogger = NULL;                        // Singleton
SysLoggerRef addedLoggers[SYSLOG_MAX_NBR_OF_LOGGERS]; // Added loggers

getLoggerInstFn pGetLoggerInstFn = &getLoggerInst;
addLoggerFn     pAddLoggerFn     = &addLogger;

UnixInitSysRef initSystemRef = NULL;

/*********************************************************************************************
 *        ILog Interface Realization
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
static void log_msg(ILogImplRef myImpl, LogLevel_t logLevel, const char * msg, va_list args)
{
    SysLoggerRef me = (SysLoggerRef)myImpl;
    assert(me != NULL);

    char logMsg[SYSLOG_MAX_MSG_LENGTH];
    vsnprintf(logMsg, SYSLOG_MAX_MSG_LENGTH, msg, args);

    //output to syslog if activated
    if(me->useSyslog)
    {
        // NOTE: - syslog() adds the facility string, we do not need to add it here
        //       - Opening the syslog facility will forward the message to the right log file
        //       - If multiple log facilities are used in one process, the syslog connection has to
        //         be opened and closed before sending messages. This is pretty slow and should
        //         only be used for low traffic loggers such as POC1..3, FGU and iDevices and
        //         NOT for tracing and heavy debug loggers!

        static SysLogFacility_t openedFacility = SYSLOG_FACILITY_MAX;
        if(openedFacility != me->logFacility)
        {
            // re-open the logger in case it changed
            openedFacility = me->logFacility;
            if(me->useSyslog)
            {
                closelog();
                openlog(SYSLOG_FACILITY_NAME[me->logFacility], 0, SYSLOG_FACILITY[me->logFacility]);
            }
        }
        // introduced to change logging device (syslogger or journald):
        IUnixInitSys_logMessage(sysLogger->initSysItfRef, (int32_t)logLevel, (int32_t)SYSLOG_FACILITY[me->logFacility],
                                (const char*)me->loggerName, (const char*)logMsg);
    }
    //output to stdout if activated
    if(me->useStdout)
    {
        printf("[%s]\t%s\t%s: %s\n", SYSLOG_LEVEL_NAME[(int)logLevel], sysLogger->appName, me->loggerName, logMsg);
    }
}

/*********************************************************************************************
 *        LOCAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
static ILogRef getLoggerByName(const char* name, SysLogFacility_t logFacility)
{
    int idx = 0;
    if((name == NULL) || (strlen(name) == 0))
    {
        Throw(EX_INVALID_ARG, "Logger name invalid");
    }

    for(;(addedLoggers[idx] != NULL) && (idx < SYSLOG_MAX_NBR_OF_LOGGERS); idx++)
    {
        bool isLoggerRegistered = (strcmp(addedLoggers[idx]->loggerName, name) == 0);
        if( isLoggerRegistered && (logFacility == addedLoggers[idx]->logFacility) )
        {
            return addedLoggers[idx]->myLogItf;
        }
    }
    return NULL;
}
//--------------------------------------------------------------------------------------------
static int32_t getNextFreeLoggerIndex(void)
{
    int32_t idx = 0;
    for(; addedLoggers[idx] != NULL && idx < SYSLOG_MAX_NBR_OF_LOGGERS; idx++);
    if(idx >= SYSLOG_MAX_NBR_OF_LOGGERS)
    {
        Throw(EX_LIMIT_CHECK, "No free loggers available, maximum count (%u) reached!",
                              (unsigned int)SYSLOG_MAX_NBR_OF_LOGGERS);
    }
    return idx;
}
//--------------------------------------------------------------------------------------------
static SysLoggerRef createLogger(const char * name, SysLogFacility_t logFacility,
                                 LogLevel_t maxLogLevel, bool useSyslog, bool useStdout, bool isSysLogger)
{
    SysLoggerRef me = (SysLoggerRef)calloc(1, sizeof(SysLoggerPrivate_t));
    if(me == NULL)
    {
        Throw(EX_OUT_OF_MEMORY, "Memory for logger instance (%s) could not be allocated", name);
    }
    me->loggerName = (char *)calloc(1, strlen(name) + 1);
    if(me->loggerName == NULL)
    {
        Throw(EX_OUT_OF_MEMORY, "Memory for logger name (%s) could not be allocated", name);
    }
    strcpy(me->loggerName, name);
    me->logFacility = logFacility;
    me->maxLogLevel = maxLogLevel;
    me->useSyslog = useSyslog;
    me->useStdout = useStdout;
    me->myLogItf = ILog_ctor(me, isSysLogger, maxLogLevel, log_msg);
    return me;
}
//--------------------------------------------------------------------------------------------
static void destroyLogger(SysLoggerRef me)
{
    if(me == NULL)
    {
        return;
    }
    ILog_destroy(me->myLogItf);
    free(me->loggerName);
    free(me);
}
//--------------------------------------------------------------------------------------------
static void addLoggersFromCmdLine(SysLoggersCmdLine_t* cfgRef)
{
    int i = 0;
    assert(sysLogger != NULL);

    // setup main system logger instance
    sysLogger->useStdout = cfgRef->useStdout;
    sysLogger->useSyslog = cfgRef->useSysLog;
    sysLogger->maxLogLevel = cfgRef->maxLogLevel;
    ILog_setMaxLogLevel(sysLogger->myLogItf, sysLogger->maxLogLevel);

    for(; (i < SYSLOG_MAX_NBR_OF_LOGGERS) && (cfgRef->loggers[i].loggerName[0] != '\0'); i++)
    {
        if(strcmp(cfgRef->loggers[i].loggerName, SYSLOGGER_NAME) == 0)
        {
            // default system logger name was configured on command line
            sysLogger->maxLogLevel = cfgRef->loggers[i].maxLogLevel;
            ILog_setMaxLogLevel(sysLogger->myLogItf, sysLogger->maxLogLevel);
        }
        else
        {
            SysLogger_addLogger(cfgRef->loggers[i].loggerName, sysLogger->logFacility,
                                cfgRef->loggers[i].maxLogLevel, sysLogger->useSyslog, sysLogger->useStdout);
        }
    }
}


//--------------------------------------------------------------------------------------------
ILogRef getLoggerInst(const char * name, bool addIfMissing)
{
    if(addIfMissing)
    {
        return SysLogger_addLogger(name, sysLogger->logFacility, sysLogger->maxLogLevel, sysLogger->useSyslog, sysLogger->useStdout);
    }
    ILogRef loggerRef = getLoggerByName(name, sysLogger->logFacility);
    if(loggerRef == NULL)
    {
        Throw(EX_NOT_FOUND, "Logger with name '%s' was not found", name);
    }
    return loggerRef;
}
//--------------------------------------------------------------------------------------------
ILogRef addLogger(const char * name, SysLogFacility_t logFacility, LogLevel_t maxLogLevel, bool useSyslog, bool useStdout)
{
    assert(sysLogger != NULL);

    if((name == NULL) || (strlen(name) == 0))
    {
        Throw(EX_INVALID_ARG, "Logger cannot be added, name invalid");
    }

    ILogRef loggerRef = getLoggerByName(name, logFacility);
    if(loggerRef != NULL)
    {
        // Found an already registered logger
        return loggerRef;
    }

    // Not existing, create one
    int freeIdx = getNextFreeLoggerIndex();
    addedLoggers[freeIdx] = createLogger(name, logFacility, maxLogLevel, useSyslog, useStdout, false);
    return addedLoggers[freeIdx]->myLogItf;
}


//--------------------------------------------------------------------------------------------
void setupLoggingSystem(SysLoggerRef me)
{
    assert(me != NULL);
    initSystemRef = UnixInitSys_create(me->myLogItf, me->appName, true);
    sysLogger->initSysItfRef = UnixInitSys_getItfRef(initSystemRef);
}
//--------------------------------------------------------------------------------------------
void destroyLoggingSystem(SysLoggerRef me)
{
    UnixInitSys_destroy(initSystemRef);
    initSystemRef = NULL;
    if(me->useSyslog)
    {
        closelog();
    }
}

/*********************************************************************************************
 *        GLOBAL FUNCTIONS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
SysLoggerRef SysLogger_create(const char* appName, LogLevel_t maxLogLevel)
{
    // the syslogger is a singleton and create shall therefore only be called once
    if(sysLogger != NULL)
    {
        return sysLogger;
    }
    assert(appName != NULL);
    memset(addedLoggers, 0, sizeof(SysLoggerRef) * SYSLOG_MAX_NBR_OF_LOGGERS);
    sysLogger = createLogger(SYSLOGGER_NAME, SYSLOG_FACILITY_ROUTER, maxLogLevel, true, false, true);
    sysLogger->appName = appName;
    setupLoggingSystem(sysLogger);
    return sysLogger;
}
//--------------------------------------------------------------------------------------------
void SysLogger_destroy(SysLoggerRef me)
{
    if(me == NULL)
    {
        return;
    }
    assert(me == sysLogger);
    int i;
    for(i = 0; i < SYSLOG_MAX_NBR_OF_LOGGERS; i++)
    {
        if(addedLoggers[i] == NULL)
        {
            break;
        }
        destroyLogger(addedLoggers[i]);
        addedLoggers[i] = NULL;
    }
    destroyLoggingSystem(sysLogger);
    destroyLogger(me);
    sysLogger = NULL;
}
//--------------------------------------------------------------------------------------------
void SysLogger_addCommandLineLoggers(SysLoggersCmdLine_t* cmdLineCfgRef)
{
    if(cmdLineCfgRef == NULL)
    {
        return;
    }
    if(sysLogger == NULL)
    {
        Throw(EX_MISSING, "Cannot add command line loggers before the system logger is created");
    }
    addLoggersFromCmdLine(cmdLineCfgRef);
}

//--------------------------------------------------------------------------------------------
ILogRef SysLogger_getLoggerItf(SysLoggerRef me)
{
    assert(me != NULL);
    return me->myLogItf;
}
//--------------------------------------------------------------------------------------------
ILogRef SysLogger_getLoggerInstance(const char* name, bool addIfMissing)
{
    return pGetLoggerInstFn(name, addIfMissing);
}
//--------------------------------------------------------------------------------------------
ILogRef SysLogger_addLogger(const char* name, SysLogFacility_t logFacility, LogLevel_t maxLogLevel, bool useSyslog, bool useStdout)
{
    return pAddLoggerFn(name, logFacility, maxLogLevel, useSyslog, useStdout);
}
//--------------------------------------------------------------------------------------------
void SysLogger_setMaxLogLevel(SysLoggerRef me, LogLevel_t maxLogLevel)
{
    assert(me != NULL);
    me->maxLogLevel = maxLogLevel;
    ILog_setMaxLogLevel(me->myLogItf, me->maxLogLevel);
}
//--------------------------------------------------------------------------------------------
LogLevel_t SysLogger_getMaxLogLevel(SysLoggerRef me)
{
    assert(me != NULL);
    return me->maxLogLevel;
}



/*********************************************************************************************
 *        FOR UNIT TESTS ONLY
 *********************************************************************************************/
void SysLogger_setupTestFunctions(getLoggerInstFn getLoggerFnRef, addLoggerFn addLoggerFnRef)
{
    pGetLoggerInstFn = (getLoggerFnRef == NULL) ? &getLoggerInst : getLoggerFnRef;
    pAddLoggerFn     = (addLoggerFnRef == NULL) ? &addLogger : addLoggerFnRef;
}


#ifdef __cplusplus
}
#endif

/************************************END OF FILE*********************************************/
