/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * Daemonizer MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : daemonizer.c
 * CREATION DATE : Apr 2, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <sys/types.h>   // for fork, setsid, getpid
#include <unistd.h>      // for fork, setsid
#include <errno.h>
#include <stdio.h>       // for fclose
#include <sys/stat.h>    // for open
#include <fcntl.h>       // for open

#include <nmapp/types/nm-types.h>
#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include <nmapp/process/daemonizer.h>

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/
#define PID_VALUE_LEN  64

/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
pid_t Daemonizer_daemonize(const char *pidFile)
{
    assert(pidFile != NULL);
    pid_t pid = fork();
    if(pid == -1)
    {
        Throw(EX_FORK, "Could not daemonize: %s", strerror(errno));
    }
    // Child process active --> daemon:
    if(pid == 0)
    {
        // be quiet on the command line output and set own session
        fclose(stdout);
        fclose(stderr);
        if(setsid() < 0)
        {
            Throw(EX_FORK, "Could not create new session: %s", strerror(errno));
        }
    }
    else
    {
        // parent process active:
        Daemonizer_writePidFile(pidFile, (int)pid);
    }
    return pid;
}
//--------------------------------------------------------------------------------------------
void Daemonizer_writePidFile(const char *pidFile, int pid)
{
    assert(pidFile != NULL);
    if(pidFile[0] == '\0')
    {
        Throw(EX_INVALID_ARG, "Empty string for PID file not allowed");
    }
    char pidValue[PID_VALUE_LEN] = "";
    int  pidFd  = open(pidFile, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
    int  thePid = (pid == -1) ? (int)getpid() : pid;
    if(pidFd < 0)
    {
        Throw(EX_CREATE, "Error creating PID file %s: %s", pidFile, strerror(errno));
    }
    snprintf(pidValue, PID_VALUE_LEN, "%d\n", thePid);
    if(write(pidFd, pidValue, strlen(pidValue)) < 0)
    {
        Throw(EX_WRITE, "Cannot write PID file %s: %s", pidFile, strerror(errno));
    }
    close(pidFd);
}
//--------------------------------------------------------------------------------------------
void Daemonizer_removePidFile(const char *pidFile)
{
    assert(pidFile != NULL);
    if(remove(pidFile) != 0)
    {
        Throw(EX_REMOVE, "Cannot remove PID file %s: %s", pidFile, strerror(errno));
    }
}

