/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * temperature MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : temperature.c
 * CREATION DATE : Apr 8, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdlib.h>          // for free

#include <nmapp/types/nm-types.h>
#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include <nmapp/conversion/conversion.h>

#include "temperature-private.h"

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/





//--------------------------------------------------------------------------------------------
void setTemperatureName(TemperatureRef me)
{
    assert(me != NULL);
    switch(me->sensorId)
    {
        case TEMP_SYS:
            snprintf(me->sensorName, TEMP_MAX_BUFFER_SIZE, "%s", TEMP_SYS_NAME);
            break;
        case TEMP_ENV:
            snprintf(me->sensorName, TEMP_MAX_BUFFER_SIZE, "%s", TEMP_ENV_NAME);
            break;
        case TEMP_NONE:
            snprintf(me->sensorName, TEMP_MAX_BUFFER_SIZE, "%s", TEMP_NONE_NAME);
            break;
        case TEMP_MAX:
        default:
            Throw(EX_INVALID_ARG, "Unknown temperature ID detected: %d", (int)me->sensorId);
    }
}

/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
TemperatureRef Temperature_create(TemperatureId_t sensorId, IFileOperationRef fileOpRef, const char* sensorPath, ILogRef logRef)
{
    assert(fileOpRef != NULL);
    TemperatureRef me = (TemperatureRef)calloc(1, sizeof(TemperaturePrivate_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Cannot create temperature instance");
    }
    me->logRef      = logRef;
    me->sensorId    = sensorId;
    me->fileOpRef   = fileOpRef;
    me->sensorPath  = sensorPath;
    me->wasFileOpen = IFileOperation_isFileOpen(me->fileOpRef);
    setTemperatureName(me);
    if(!me->wasFileOpen)
    {
        IFileOperation_open(me->fileOpRef, me->sensorPath, "r");
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "%s temperature successfully setup", me->sensorName);
    return me;
}
//--------------------------------------------------------------------------------------------
void Temperature_destroy(TemperatureRef me)
{
    if(me != NULL)
    {
        ILog_msg(me->logRef, LOG_LEVEL_DBG, "Destroying %s temperature instance", me->sensorName);
        if(!me->wasFileOpen)
        {
            IFileOperation_close(me->fileOpRef);
        }
        free(me);
    }
}
//--------------------------------------------------------------------------------------------
float64_t Temperature_read(TemperatureRef me)
{
    assert(me != NULL);

    char tempString[TEMP_MAX_BUFFER_SIZE];
    memset(tempString, 0, TEMP_MAX_BUFFER_SIZE);
    IFileOperation_readFromFile(me->fileOpRef, &tempString[0], (TEMP_MAX_BUFFER_SIZE - 1), 0);
    size_t tempStrLen = strlen(&tempString[0]);
    assert(tempStrLen > 0);
    char lastTempStrChar = tempString[tempStrLen-1];
    tempString[tempStrLen-1] = (lastTempStrChar == '\n') ? '\0' : lastTempStrChar;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "Read %s temperature: rawTemp=%s", me->sensorName, tempString);
    float64_t temperatureC = str2float(tempString);
    temperatureC = (float64_t)(temperatureC/1000);
    return temperatureC;
}
