/*********************************************************************************************
 * (c) COPYRIGHT 2021 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * UnixInitSys MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : UnixInitSys.c
 * CREATION DATE : Jan 18, 2021
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdlib.h>                    // for calloc, free, ...
#include <errno.h>
#include <string.h>
#include <syslog.h>

#include <nmapp/exception/exception.h>
#include <nmapp/container/uthash.h>

#if(defined(__USE_INIT_SYS_DUMMY))

typedef struct sd_bus { int dummy; } sd_bus;
typedef struct sd_event { int dummy; } sd_event;
typedef struct sd_bus_error { int dummy; } sd_bus_error;
typedef struct sd_bus_slot { int dummy; } sd_bus_slot;
typedef struct sd_bus_message { int dummy; } sd_bus_message;
typedef int  (*sd_bus_message_handler_t)(sd_bus_message *m, void *userdata, sd_bus_error *ret_error);

int          logToJournalDummy(const char *format, ...) { return 0; }
int          createEventDefaultDummy(sd_event **e) { return 0; }
int          getEventFdDummy(sd_event *e) { return 42; /* returning just a magic number (stdout, stderr, stdin, ..) */ }
int          prepareEventDummy(sd_event *e) { return 0; }
int          waitEventDummy(sd_event *e, uint64_t usec) { return 0; }
int          dispatchEventDummy(sd_event *e) { return 0; }
sd_event*    unrefEventDummy(sd_event *e) { return NULL; }
int          exitEventDummy(sd_event *e, int code) { return code; }
int          sdEventLoopDummy(sd_event *e) { return 0; }
int          openSdSystemBusDummy(sd_bus **ret) { return 0; }
void         closeSdBusDummy(sd_bus *bus) {}
int          attachEventToSdBusDummy(sd_bus *bus, sd_event *e, int priority) { return 0; }
int          matchSignalOnSdBusDummy(sd_bus *bus, sd_bus_slot **slot,
                                     const char *sender, const char *path, const char *interface, const char *member,
                                     sd_bus_message_handler_t callback, void *userdata) { return 0; }
const char*  getDestinationFromSdBusMsg(sd_bus_message *message) { return NULL; }
const char*  getPathFromSdBusMsg(sd_bus_message *message) { return NULL; }
const char*  getInterfaceFromSdBusMsg(sd_bus_message *message) { return NULL; }
const char*  getMemberFromSdBusMsg(sd_bus_message *m) { return NULL; }
int          readBasicMsgFromSdBus(sd_bus_message *m, char type, void *p) { return 0; }
int          getSdBusPropertyStringDummy(sd_bus *bus, const char *destination, const char *path, const char *interface,
                                       const char *member, sd_bus_error *ret_error, char **ret) { return 0; }
sd_bus_slot* unrefSdBusSlotDummy(sd_bus_slot *slot) { return NULL; }

#else

#include <systemd/sd-journal.h>
#include <systemd/sd-bus.h>
#include <systemd/sd-event.h>

#endif


#include "UnixInitSys-private.h"



/*********************************************************************************************
 *        EXTERNALS
 *********************************************************************************************/

const char* SYSLOG_LEVEL_NAME[LOG_LEVEL_MAX] =
{
    [LOG_LEVEL_EMCY ] = "EMCY",
    [LOG_LEVEL_ALERT] = "ALERT",
    [LOG_LEVEL_CRIT ] = "CRIT",
    [LOG_LEVEL_ERR  ] = "ERROR",
    [LOG_LEVEL_WARN ] = "WARN",
    [LOG_LEVEL_NOTE ] = "NOTE",
    [LOG_LEVEL_INFO ] = "INFO",
    [LOG_LEVEL_DBG  ] = "DEBUG",
};



/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/

typedef struct __UnixInitSysBusPrivate
{
    sd_bus*                            bus;
    sd_bus_error                       busErr;
    sd_event*                          sdEventRef;
    UnixInitSysDBusSignalItemCallback  evtCallbackFn;
    void*                              evtCallbackContext;
} UnixInitSysBusPrivate_t;




//--------------------------------------------------------------------------------------------
void logMessage(IUnixInitSysImplRef implRef, int32_t logPriority, int32_t facility, const char* loggerName, const char* msg)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    if(me->useSysLog)
    {
        syslog(logPriority | facility, SYSLOG_MSG_FORMAT_STR, SYSLOG_LEVEL_NAME[logPriority], me->appName, loggerName, msg);
        return;
    }
    if(me->logToJournal("PRIORITY=%d", logPriority, "SYSLOG_FACILITY=%d", facility, "SYSLOG_IDENTIFIER=%s", loggerName,
                       "_SYSTEMD_UNIT=%s", me->appName, "MESSAGE=%s", msg, NULL) != 0)
    {
        Throw(EX_WRITE, "log message to journal: %s", strerror(errno));
    }

}


//--------------------------------------------------------------------------------------------
void initDBus(IUnixInitSysImplRef implRef)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "init operating state monitor...");
    int r = me->sdbusSysOpen(&me->busRef->bus);
    if(r < 0)
    {
        Throw(EX_OPEN, "Failed to open sd bus: %s", strerror(-r));
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "bus openend for operating state monitoring");
    r = me->sdEventDefault(&me->busRef->sdEventRef);
    if(r < 0)
    {
        Throw(EX_CREATE, "Failed creating sd bus event: %s", strerror(-r));
    }
    r = me->sdbusAttachEvent(me->busRef->bus, me->busRef->sdEventRef, 0);
    if(r < 0)
    {
        Throw(EX_CREATE, "Failed attaching the sd bus event to the bus: %s", strerror(-r));
    }

    ILog_msg(me->logRef, LOG_LEVEL_DBG, "attached event on bus for operating state monitoring");
    me->fdEvent = me->sdEventGetFd(me->busRef->sdEventRef);
}
//--------------------------------------------------------------------------------------------
void deinitDBus(IUnixInitSysImplRef implRef)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    me->sdEventUnref(me->busRef->sdEventRef);
    me->sdbusClose(me->busRef->bus);
    me->fdEvent = -1;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "de-initialized operating state monitor");
}
//--------------------------------------------------------------------------------------------
int getEventFd(IUnixInitSysImplRef implRef)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    return me->fdEvent;
}
//--------------------------------------------------------------------------------------------
void registerDBusSignalCallback(IUnixInitSysImplRef implRef, UnixInitSysDBusSignalItemCallback cbFn, void* pCbCtx)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    me->busRef->evtCallbackFn = cbFn;
    me->busRef->evtCallbackContext = pCbCtx;
}
//--------------------------------------------------------------------------------------------
void activateDBusListener(IUnixInitSysImplRef implRef)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    int sdRet = me->sdEventPrepare(me->busRef->sdEventRef);     // prepare once for getting the events
    if(sdRet < 0)
    {
        Throw(EX_SYNC, "dbus listener activation failed: %s", strerror(-sdRet));
    }
}
//--------------------------------------------------------------------------------------------
void processDBusEvents(IUnixInitSysImplRef implRef)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "processing dbus events");
    // The event is actually already ready but we need to call wait for the dbus state
    //  machine and then we need to call for dispatching the events
    int sdRet = me->sdEventWait(me->busRef->sdEventRef, 0);
    if(sdRet < 0)
    {
        Throw(EX_SYNC, "sd event wait failed: %s", strerror(-sdRet));
    }
    sdRet = me->sdEventDispatch(me->busRef->sdEventRef);
    if(sdRet < 0)
    {
        Throw(EX_SYNC, "sd event dispatch failed: %s", strerror(-sdRet));
    }
    sdRet = me->sdEventPrepare(me->busRef->sdEventRef);         // prepare again for the next event
    if(sdRet < 0)
    {
        Throw(EX_SYNC, "sd event prepare failed: %s", strerror(-sdRet));
    }
}
//--------------------------------------------------------------------------------------------
static int dbusSignalCallback(sd_bus_message* dbusMsg, void* pCtx, sd_bus_error* retError)
{
    UnixInitSysRef me = (UnixInitSysRef)pCtx;
    const char *item        = NULL;
    const char* signalName  = me->sdbusMsgGetMember(dbusMsg);
    const char* destination = me->sdbusMsgGetDest(dbusMsg);
    const char* path        = me->sdbusMsgGetPath(dbusMsg);
    const char* itf         = me->sdbusMsgGetItf(dbusMsg);
    me->sdbusMsgReadBasic(dbusMsg, 's', (void*)&item);

    if(me->busRef->evtCallbackFn != NULL)
    {
        ILog_msg(me->logRef, LOG_LEVEL_DBG, "Calling dbus event callback for signal/item=%s/%s", signalName, item);
        me->busRef->evtCallbackFn(me->busRef->evtCallbackContext, destination, path, itf, signalName, item);
        return 0;
    }
    ILog_msg(me->logRef, LOG_LEVEL_INFO, "dbus signal callback without callback: name=%s, item=%s", signalName, item);
    return 0;
}
//--------------------------------------------------------------------------------------------
void* registerDBusSignalMatch(IUnixInitSysImplRef implRef, const char* dbusService, const char* dbusObjPath,
                              const char* dbusItfName, const char* dbusSignalName)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    sd_bus_slot* pBusSlot = NULL;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "register dbus signal match for signal=%s", dbusSignalName);
    int regRet = me->sdbusMatchSignal(me->busRef->bus, &pBusSlot, dbusService, dbusObjPath, dbusItfName,
                                      dbusSignalName, &dbusSignalCallback, (void*)me);
    if(regRet < 0)
    {
        Throw(EX_ADD, "Registering dbus signal=%s match: %s", dbusSignalName, strerror(-regRet));
    }
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "successfully registered dbus signal match, ref=%p", pBusSlot);
    return (void*)pBusSlot;
}
//--------------------------------------------------------------------------------------------
void unregisterDBusSignalMatch(IUnixInitSysImplRef implRef, void* signalMatchRef)
{
    assert(implRef != NULL);
    UnixInitSysRef me = (UnixInitSysRef)implRef;
    sd_bus_slot* pBusSlot = (sd_bus_slot*)signalMatchRef;
    me->sdbusUnrefSlot(pBusSlot);
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "unregister dbus signal match with ref=%p", signalMatchRef);
}
//--------------------------------------------------------------------------------------------
const char* getOperatingState(IUnixInitSysImplRef implRef)
{
    assert(implRef != NULL);

    const char* bus_dest = "org.freedesktop.systemd1";
    const char* bus_path = "/org/freedesktop/systemd1";
    const char* bus_itf = "org.freedesktop.systemd1.Manager";

    UnixInitSysRef me = (UnixInitSysRef)implRef;
    char* returnedOpState = NULL;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "getting operating state from dbus...");
    int r = me->sdbusGetPropertyStr(me->busRef->bus, bus_dest, bus_path, bus_itf, "SystemState", &me->busRef->busErr, &returnedOpState);
    if(r < 0)
    {
        Throw(EX_READ, "Failed querying the system state: %s", strerror(-r));
    }
    snprintf(&me->opState[0], UNIX_INIT_SYS_BUF_SIZE, "%s", (const char*)returnedOpState);
    free(returnedOpState);
    return (const char*)(&me->opState[0]);
}


/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
UnixInitSysRef UnixInitSys_create(ILogRef logRef, const char* appName, bool useSysLog)
{
    UnixInitSysRef me = (UnixInitSysRef)calloc(1, sizeof(UnixInitSysPrivate_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Cannot create unix init-sys instance");
    }
    UnixInitSysBusRef busRef = (UnixInitSysBusRef)calloc(1, sizeof(UnixInitSysBusPrivate_t));
    if(busRef == NULL)
    {
        Throw(EX_CREATE, "Cannot create an instance of unix inis-sys bus");
    }
    me->logRef    = logRef;
    me->appName   = appName;
    me->useSysLog = useSysLog;
    me->fdEvent   = -1;
    me->busRef    = busRef;

#if(defined(__USE_INIT_SYS_DUMMY))
    me->logToJournal        = &logToJournalDummy;
    me->sdEventDefault      = &createEventDefaultDummy;
    me->sdEventGetFd        = &getEventFdDummy;
    me->sdEventPrepare      = &prepareEventDummy;
    me->sdEventDispatch     = &dispatchEventDummy;
    me->sdEventWait         = &waitEventDummy;
    me->sdEventUnref        = &unrefEventDummy;
    me->sdbusSysOpen        = &openSdSystemBusDummy;
    me->sdbusClose          = &closeSdBusDummy;
    me->sdbusAttachEvent    = &attachEventToSdBusDummy;
    me->sdbusMatchSignal    = &matchSignalOnSdBusDummy;
    me->sdbusMsgGetDest     = &getDestinationFromSdBusMsg;
    me->sdbusMsgGetPath     = &getPathFromSdBusMsg;
    me->sdbusMsgGetItf      = &getInterfaceFromSdBusMsg;
    me->sdbusMsgGetMember   = &getMemberFromSdBusMsg;
    me->sdbusMsgReadBasic   = &readBasicMsgFromSdBus;
    me->sdbusGetPropertyStr = &getSdBusPropertyStringDummy;
    me->sdbusUnrefSlot      = &unrefSdBusSlotDummy;
#else
    me->busRef->busErr      = SD_BUS_ERROR_NULL;

    me->logToJournal        = &sd_journal_send;
    me->sdEventDefault      = &sd_event_default;
    me->sdEventGetFd        = &sd_event_get_fd;
    me->sdEventPrepare      = &sd_event_prepare;
    me->sdEventDispatch     = &sd_event_dispatch;
    me->sdEventWait         = &sd_event_wait;
    me->sdEventUnref        = &sd_event_unref;
    me->sdbusSysOpen        = &sd_bus_open_system;
    me->sdbusClose          = &sd_bus_close;
    me->sdbusAttachEvent    = &sd_bus_attach_event;
    me->sdbusMatchSignal    = &sd_bus_match_signal;
    me->sdbusMsgGetDest     = &sd_bus_message_get_destination;
    me->sdbusMsgGetPath     = &sd_bus_message_get_path;
    me->sdbusMsgGetItf      = &sd_bus_message_get_interface;
    me->sdbusMsgGetMember   = &sd_bus_message_get_member;
    me->sdbusMsgReadBasic   = &sd_bus_message_read_basic;
    me->sdbusGetPropertyStr = &sd_bus_get_property_string;
    me->sdbusUnrefSlot      = &sd_bus_slot_unref;
#endif
    me->itf       = IUnixInitSys_ctor(me,
                                      &logMessage,
                                      &initDBus,
                                      &deinitDBus,
                                      &getEventFd,
                                      &registerDBusSignalCallback,
                                      &activateDBusListener,
                                      &processDBusEvents,
                                      &registerDBusSignalMatch,
                                      &unregisterDBusSignalMatch,
                                      &getOperatingState);
    return me;
}
//--------------------------------------------------------------------------------------------
void UnixInitSys_destroy(UnixInitSysRef me)
{
    if(me != NULL)
    {
        IUnixInitSys_destroy(me->itf);
        free(me->busRef);
        free(me);
    }
}
//--------------------------------------------------------------------------------------------
IUnixInitSysRef UnixInitSys_getItfRef(UnixInitSysRef me)
{
    assert(me != NULL);
    return me->itf;
}

