/*********************************************************************************************
 * (c) COPYRIGHT 2021 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * system-state MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : system-state.c
 * CREATION DATE : Jan 15, 2021
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <stdlib.h>                    // for calloc, free, ...
#include <errno.h>
#include <string.h>
#include <nmapp/types/nm-types.h>
#include <nmapp/exception/exception.h>
#include <nmapp/sys/assertion.h>
#include "system-state-private.h"

/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/


/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/

//--------------------------------------------------------------------------------------------
SystemStateRef SystemState_create(ILogRef logRef, IUnixInitSysRef initSysItfRef)
{
    SystemStateRef me = (SystemStateRef)calloc(1, sizeof(SystemStatePrivate_t));
    if(me == NULL)
    {
        Throw(EX_CREATE, "Cannot create system state instance");
    }
    me->logRef     = logRef;
    me->initSysItf = initSysItfRef;
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "system state instance successfully created");
    return me;
}
//--------------------------------------------------------------------------------------------
void SystemState_destroy(SystemStateRef me)
{
    if(me != NULL)
    {
        ILog_msg(me->logRef, LOG_LEVEL_DBG, "Destroying system state instance");
        SystemState_deinit(me);
        free(me);
    }
}
//--------------------------------------------------------------------------------------------
void SystemState_init(SystemStateRef me)
{
    assert(me != NULL);
    IUnixInitSys_initDBus(me->initSysItf);
}
//--------------------------------------------------------------------------------------------
void SystemState_deinit(SystemStateRef me)
{
    assert(me != NULL);
    IUnixInitSys_deinitDBus(me->initSysItf);
}
//--------------------------------------------------------------------------------------------
int SystemState_getOperatingStateFileDescriptor(SystemStateRef me)
{
    return IUnixInitSys_getEventFileDescriptor(me->initSysItf);
}
//--------------------------------------------------------------------------------------------
void SystemState_operatingStateChangeCallback(void *pCtx)
{
    SystemStateRef me = (SystemStateRef)pCtx;
    const char  *opState = IUnixInitSys_getOperatingState(me->initSysItf);
    ILog_msg(me->logRef, LOG_LEVEL_DBG, "system state: %s", opState);
    bool isInitializing = (strncmp(opState, "initializing", 16) == 0);
    bool isStarting = (strncmp(opState, "starting", 16) == 0);
    if(isStarting || isInitializing)
    {
        return;
    }

    me->isSystemUp = true;
    SystemState_deinit(me);
}
//--------------------------------------------------------------------------------------------
bool SystemState_isSystemUp(SystemStateRef me)
{
    assert(me != NULL);
    return me->isSystemUp;
}


