/*********************************************************************************************
 * (c) COPYRIGHT 2020 by NetModule AG, Switzerland.  All rights reserved.
 * -------------------------------------------------------------------------------------------
 *
 * system MODULE
 *
 * -------------------------------------------------------------------------------------------
 *
 * FILENAME      : system.c
 * CREATION DATE : Apr 1, 2020
 * TARGET        : embedded Linux
 *
 *********************************************************************************************/

#include <errno.h>
#include <stdio.h>         // for available RAM (fopen, fgets)
#include <stdlib.h>        // for get load averages
#include <sys/sysinfo.h>   // for uptime, load average
#include <sys/statvfs.h>   // for disk usage (statvfs)

#include <nmapp/exception/exception.h>
#include "system-private.h"


/*********************************************************************************************
 *        LOCALS
 *********************************************************************************************/
#define  PROC_READBUFFER_SIZE             128              //!< Buffer size when reading from /proc/xyz

SystemPrivat_t localMe = {
        .availableRamKb   = 0,
        .diskUsageMB      = {0.0},
        .cpuLoadTripple   = {0.0},
        .pGetDiskUsage    = &getDiskUsageInfo,
        .pGetAvailableRAM = &getAvailableRAM,
        .pGetCpuLoad      = &getCpuLoad,
        .pGetUptime       = &getUptime
};




//--------------------------------------------------------------------------------------------
float64_t* getDiskUsageInfo(const char* path)
{
    struct statvfs usageInfo;
    if(statvfs(path, &usageInfo) != 0)
    {
        Throw(EX_READ, "Getting disk usage (path=%s): %s", path, strerror(errno));
    }
    localMe.diskUsageMB[SYS_DISKUSAGE_TOTAL]     = (float64_t)(usageInfo.f_blocks * usageInfo.f_frsize) / DIGITAL_MEGA;
    localMe.diskUsageMB[SYS_DISKUSAGE_AVAILABLE] = (float64_t)(usageInfo.f_bfree  * usageInfo.f_bsize)  / DIGITAL_MEGA;
    localMe.diskUsageMB[SYS_DISKUSAGE_FREE]      = (float64_t)(usageInfo.f_favail * usageInfo.f_bsize)  / DIGITAL_MEGA;
    localMe.diskUsageMB[SYS_DISKUSAGE_USED]      = (float64_t)(localMe.diskUsageMB[SYS_DISKUSAGE_TOTAL] - localMe.diskUsageMB[SYS_DISKUSAGE_AVAILABLE]);
    return &localMe.diskUsageMB[0];
}
//--------------------------------------------------------------------------------------------
uint32_t getAvailableRAM(void)
{
    const char *memInfoFilepath = "/proc/meminfo";
    char  line[PROC_READBUFFER_SIZE] = "";
    FILE *pFile = NULL;
    bool  isFound = false;
    localMe.availableRamKb = 0;

    pFile = fopen(memInfoFilepath, "r");
    if(pFile == NULL)
    {
        Throw(EX_OPEN, "Cannot open %s: %s", memInfoFilepath, strerror(errno));
    }
    while(fgets(&line[0], PROC_READBUFFER_SIZE, pFile) != NULL)
    {
        if(sscanf(line, "MemAvailable: %d kB", (int*)&localMe.availableRamKb) == 1)
        {
            isFound = true;
            break;
        }
    }
    fclose(pFile);
    if(!isFound)
    {
        Throw(EX_NOT_FOUND, "Available RAM in %s not found", memInfoFilepath);
    }
    return localMe.availableRamKb;
}
//--------------------------------------------------------------------------------------------
float64_t* getCpuLoad(void)
{
    if (getloadavg((double*)&localMe.cpuLoadTripple, SYS_CPULOAD_MAX) < 0)
    {
        Throw(EX_READ, "Geting cpu load triple: %s", strerror(errno));
    }
    return &localMe.cpuLoadTripple[0];
}
//--------------------------------------------------------------------------------------------
uint64_t getUptime(void)
{
    struct sysinfo systemInfo;
    memset(&systemInfo, 0, sizeof(systemInfo));
    if (sysinfo(&systemInfo) != 0)
    {
        Throw(EX_READ, "Getting system info for uptime: %s", strerror(errno));
    }
    return (uint64_t)(systemInfo.uptime);
}


/*********************************************************************************************
 *        GLOBALS
 *********************************************************************************************/
//--------------------------------------------------------------------------------------------
float64_t* System_getDiskUsageInfo(const char* path)
{
    return localMe.pGetDiskUsage(path);
}
//--------------------------------------------------------------------------------------------
uint32_t System_getAvailableRAM(void)
{
    return localMe.pGetAvailableRAM();
}
//--------------------------------------------------------------------------------------------
float64_t* System_getCpuLoad(void)
{
    return localMe.pGetCpuLoad();
}
//--------------------------------------------------------------------------------------------
uint64_t System_getUptime(void)
{
    return localMe.pGetUptime();
}


/*********************************************************************************************
 *        FOR UNIT TESTS ONLY
 *********************************************************************************************/
void System_setupSystemCalls(getDiskUsageFn getDiskUsageRef,
                             getAvailableRAMFn getAvailableRamRef,
                             getCpuLoadFn getCpuLoadRef,
                             getUptimeFn getUptimeRef)
{
    localMe.pGetDiskUsage    = (getDiskUsageRef == NULL)    ? &getDiskUsageInfo : getDiskUsageRef;
    localMe.pGetAvailableRAM = (getAvailableRamRef == NULL) ? &getAvailableRAM  : getAvailableRamRef;
    localMe.pGetCpuLoad      = (getCpuLoadRef == NULL)      ? &getCpuLoad       : getCpuLoadRef;
    localMe.pGetUptime       = (getUptimeRef == NULL)       ? &getUptime        : getUptimeRef;
}
